from collections import defaultdict
from functools import partial

import geopandas as gpd
import pyproj
from geoalchemy2.shape import from_shape, to_shape
from shapely.geometry import box
from shapely.ops import transform

from ..exceptions import ProcessingError
from ..models.genetic_algorithm import GAContext, GeneticAlgorithm
from ..models.geo_data import (
    AttributeType,
    DataAttributeQuantitative,
    GeoFeature,
)


def get_smallest_bounds(geometry_extent, project_extent, proj="epsg:3857"):
    """Return the bounds of the intersection of the two geometries."""
    proj_wgs84 = pyproj.Proj(init="epsg:4326")
    proj_other = pyproj.Proj(proj)
    project_to_wgs84 = partial(
        pyproj.transform, proj_other, proj_wgs84, always_xy=True
    )
    project_to_other = partial(
        pyproj.transform, proj_wgs84, proj_other, always_xy=True
    )
    b1 = transform(project_to_wgs84, geometry_extent)
    intersection = b1.intersection(project_extent)
    bounds = transform(project_to_other, intersection).bounds
    return bounds


def ga_context_from_decision_map(decision_map, geo_size, proj="epsg:3857"):
    proj_in = pyproj.Proj(init="epsg:4326")
    proj_out = pyproj.Proj(proj)
    project = partial(pyproj.transform, proj_in, proj_out)

    geo_data = decision_map.data
    attribute = geo_data.attributes[0]  # TODO: Allow to choose the attribute
    if attribute.type not in (
        AttributeType.ordinal,
        AttributeType.quantitative,
    ):
        raise ProcessingError("The attribute is not a number or a category")
    data = defaultdict(list)
    input_features = []
    for feature in geo_data.features:
        if proj_in.is_exact_same(proj_out):
            geometry = to_shape(feature.geom)
        else:
            geometry = transform(project, to_shape(feature.geom))
        geometry = geometry.buffer(
            0
        )  # use a buffer to correct invalid geometries
        value = next(
            p.value for p in feature.properties if p.attribute is attribute
        )
        input_features.append(feature)
        data["geometry"].append(geometry)
        if attribute.type is AttributeType.ordinal:
            data["value"].append(
                len(attribute.order) - attribute.order.index(value)
            )
        elif attribute.type is AttributeType.quantitative:
            data["value"].append(value)
    project_extent = to_shape(decision_map.project.extent)
    df = gpd.GeoDataFrame(data)
    not_empty = df[~df.geometry.is_empty]
    geoms_bounds = box(*not_empty.total_bounds)
    bounds = get_smallest_bounds(geoms_bounds, project_extent, proj)
    ga_ctx = GAContext(
        df, input_features, geo_size=geo_size, bounds=bounds, proj=proj
    )
    return ga_ctx


def ga_parametrized(context, params):
    return GeneticAlgorithm(context, **params)


def zone_proposition_from_genetic_algorithm(genetic_algorithm):
    features = []
    population = genetic_algorithm.filter_best()
    attribute = DataAttributeQuantitative(name="fitness")
    value_cls = attribute.get_value_class()
    for individual in population.individuals:
        value = value_cls(value=individual.fitness, attribute=attribute)
        feature = GeoFeature(
            from_shape(individual.as_geometry(proj="epsg:4326"), srid=4326),
            properties=[value],
        )
        features.append(feature)
    return features



"""Generalize project recommendations as a processing model

Revision ID: 89755dd43f58
Revises: 339b5d6198df
Create Date: 2025-04-10 13:24:13.983551

"""

from datetime import datetime, timezone

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry
from sqlalchemy.dialects import postgresql

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "89755dd43f58"
down_revision = "339b5d6198df"
branch_labels = None
depends_on = None


old_modeltype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    name="modeltype",
)
new_modeltype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    "zone_proposition",
    name="modeltype",
)
tmp_modeltype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    "zone_proposition",
    name="_modeltype",
)

old_task_type = sa.Enum(
    "process_project_data",
    "process_zone_proposition",
    "update_stream",
    name="projecttasktype",
)
new_task_type = sa.Enum(
    "process_project_data",
    "update_stream",
    name="projecttasktype",
)
tmp_task_type = sa.Enum(
    "process_project_data",
    "update_stream",
    name="_projecttasktype",
)

old_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    name="geodatatype",
)
new_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    "global_data",
    name="geodatatype",
)
tmp_geodatatype = sa.Enum(
    "geo_data",
    "generated_geo_data",
    "proposition_geo_data",
    "global_data",
    name="_geodatatype",
)

base_data = sa.Table(
    "base_data",
    sa.MetaData(),
    sa.Column("type", new_geodatatype, nullable=True),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)

processing_model = sa.Table(
    "processing_model",
    sa.MetaData(),
    sa.Column(
        "model_type",
        new_modeltype,
        nullable=True,
    ),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("data_generator_id", sa.Integer(), nullable=False),
    sa.Column(
        "keep_overlap",
        sa.Enum("min", "max", "sum", "average", name="keepoverlap"),
        nullable=False,
    ),
    sa.Column("cut_to_extent", sa.Boolean(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("radius", sa.BigInteger(), nullable=True),
    sa.Column("majority_threshold", sa.Float(), nullable=True),
    sa.Column("rule", sa.String(), nullable=True),
    sa.Column("geo_size", sa.Float(), nullable=True),
    sa.Column("iterations", sa.Integer(), nullable=True),
    sa.Column("duration", sa.Float(), nullable=True),
    sa.Column("size", sa.Integer(), nullable=True),
    sa.Column("mutation", sa.Float(), nullable=True),
    sa.Column("children", sa.Integer(), nullable=True),
    sa.Column("filter_clusters", sa.Boolean(), nullable=True),
)

project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)

input_assoc = sa.Table(
    "data_input_association",
    sa.MetaData(),
    sa.Column("project_data_id", sa.Integer(), nullable=False),
    sa.Column("input_data_id", sa.Integer(), nullable=False),
)

zone_proposition = sa.Table(
    "zone_proposition",
    sa.MetaData(),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
)

project_task = sa.Table(
    "project_task",
    sa.MetaData(),
    sa.Column("task_id", sa.String(), nullable=True),
    sa.Column("started_at", sa.DateTime(), nullable=True),
    sa.Column("finished_at", sa.DateTime(), nullable=True),
    sa.Column("state", sa.String(), nullable=True),
    sa.Column("error_message", sa.String(), nullable=True),
    sa.Column(
        "type",
        sa.Enum(
            "process_project_data",
            "process_zone_proposition",
            name="projecttasktype",
        ),
        nullable=True,
    ),
    sa.Column(
        "params", postgresql.JSONB(astext_type=sa.Text()), nullable=True
    ),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("id", sa.Integer(), nullable=False),
)

project_root = sa.Table(
    "project_root",
    sa.MetaData(),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("project_root_data_id", sa.Integer(), nullable=True),
)


def upgrade():
    # Replace old modeltype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_modeltype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE processing_model ALTER COLUMN model_type TYPE _modeltype"
        " USING model_type::text::_modeltype"
    )
    old_modeltype.drop(op.get_bind(), checkfirst=False)
    new_modeltype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE processing_model ALTER COLUMN model_type TYPE modeltype"
        " USING model_type::text::modeltype"
    )
    tmp_modeltype.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("processing_model", schema=None) as batch_op:
        batch_op.add_column(sa.Column("geo_size", sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column("iterations", sa.Integer(), nullable=True)
        )
        batch_op.add_column(sa.Column("duration", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("size", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("mutation", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("children", sa.Integer(), nullable=True))
        batch_op.add_column(
            sa.Column("filter_clusters", sa.Boolean(), nullable=True)
        )

    # Replace all zone propositions/proposition geo data with zone proposition
    # project data
    conn = op.get_bind()

    # Initialize progress bar
    nb = conn.execute(
        sa.text("SELECT count(id) FROM zone_proposition")
    ).fetchone()[0]
    count = 0
    print("Migrating Project Recommendations")
    progress_bar(count, nb)

    zone_prop_res = conn.execute(
        sa.select(
            zone_proposition.c["data_id", "project_id"],
            base_data.c["name", "created_at", "modified_at"],
            project_root.c["project_root_data_id"],
        )
        .join(base_data, zone_proposition.c.data_id == base_data.c.id)
        .join(
            project_root,
            project_root.c.project_id == zone_proposition.c.project_id,
        )
    )
    now = datetime.now(timezone.utc)
    new_model_map = {}
    for (
        data_id,
        project_id,
        name,
        created_at,
        modified_at,
        root_id,
    ) in zone_prop_res:
        # Fetch project task that built recommendations
        task_params = conn.execute(
            sa.select(project_task.c["params"])
            .where(
                project_task.c.project_id == project_id,
                project_task.c.type == "process_zone_proposition",
                project_task.c.state == "SUCCESS",
            )
            .order_by(sa.desc(project_task.c.finished_at))
        ).fetchone()[0]
        if "iterations" not in task_params and "duration" not in task_params:
            task_params["iterations"] = 1
        if "ga_params" not in task_params:
            task_params["ga_params"] = {}
        task_params["ga_params"]["mutation"] = task_params["ga_params"].get(
            "mutation", 0.02
        )
        task_params["ga_params"]["children"] = task_params["ga_params"].get(
            "children", 60
        )
        task_params["ga_params"]["filter_clusters"] = task_params[
            "ga_params"
        ].get("filter_clusters", True)
        task_params["ga_params"]["size"] = task_params["ga_params"].get(
            "size", 80
        )
        # Create new project data for recommendations
        pdata_id = conn.execute(
            project_data.insert()
            .values(
                data_type="generator",
                project_id=project_id,
                name=name,
                created_at=created_at,
                modified_at=now,
                last_update=modified_at,
                data_id=data_id,
            )
            .returning(project_data.c["id"])
        ).fetchone()[0]
        # Create zone proposition processing model
        model_id = conn.execute(
            processing_model.insert()
            .values(
                model_type="zone_proposition",
                data_generator_id=pdata_id,
                keep_overlap="max",
                geo_size=task_params.get("geo_size", 0),
                iterations=task_params.get("iterations"),
                duration=task_params.get("duration"),
                **task_params["ga_params"],
            )
            .returning(processing_model.c["id"])
        ).fetchone()[0]
        # Use new model as active model for new project data
        conn.execute(
            project_data.update()
            .values(active_model_id=model_id)
            .where(project_data.c.id == pdata_id)
        )
        new_model_map[project_id] = model_id
        # Change type of proposition geo data as generated geo data
        conn.execute(
            base_data.update()
            .values(type="generated_geo_data")
            .where(base_data.c.id == data_id)
        )
        # Add Root as input of new project data
        conn.execute(
            input_assoc.insert().values(
                project_data_id=pdata_id, input_data_id=root_id
            )
        )
        count += 1
        progress_bar(count, nb)

    # Initialize progress bar
    nb = conn.execute(
        sa.text(
            "SELECT count(project_id) FROM project_task "
            "WHERE type = 'process_zone_proposition'"
        )
    ).fetchone()[0]
    count = 0
    print("Migrating Project Tasks")
    progress_bar(count, nb)

    recomm_task_res = conn.execute(
        sa.select(project_task.c["project_id"]).where(
            project_task.c.type == "process_zone_proposition"
        )
    )
    for (project_id,) in recomm_task_res:
        conn.execute(
            project_task.update()
            .values(
                type="process_project_data",
                params={"model_id": new_model_map.get(project_id)},
            )
            .where(
                project_task.c.project_id == project_id,
                project_task.c.type == "process_zone_proposition",
            )
        )
        count += 1
        progress_bar(count, nb)

    # Replace old geodatatype with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    old_geodatatype.drop(op.get_bind(), checkfirst=False)
    new_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_geodatatype.drop(op.get_bind(), checkfirst=False)

    # Replace old task type with new one
    tmp_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE _projecttasktype"
        " USING type::text::_projecttasktype"
    )
    old_task_type.drop(op.get_bind(), checkfirst=False)
    new_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE projecttasktype"
        " USING type::text::projecttasktype"
    )
    tmp_task_type.drop(op.get_bind(), checkfirst=False)

    op.drop_table("zone_proposition")


def downgrade():
    op.create_table(
        "zone_proposition",
        sa.Column("data_id", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column(
            "project_id", sa.INTEGER(), autoincrement=False, nullable=True
        ),
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.ForeignKeyConstraint(
            ["data_id"],
            ["base_data.id"],
            name="fk_zone_proposition_data_id_base_data",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_zone_proposition_project_id_project",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_zone_proposition"),
    )

    now = datetime.now(timezone.utc)
    conn = op.get_bind()
    nb = conn.execute(
        sa.text(
            "SELECT count(pm.id) FROM processing_model pm "
            "JOIN project_data pd ON pd.active_model_id = pm.id "
            "WHERE pm.model_type = 'zone_proposition'"
        )
    ).fetchone()[0]
    count = 0
    progress_bar(count, nb)
    active_zone_props = conn.execute(
        sa.select(
            processing_model.c["id", "data_generator_id"],
            project_data.c["data_id"],
        )
        .join(
            project_data,
            project_data.c.active_model_id == processing_model.c.id,
        )
        .where(processing_model.c.model_type == "zone_proposition")
    )
    # Iterate over all active zone proposition models
    for model_id, pdata_id, data_id in active_zone_props:
        # Get other-types of model of the project_data
        other_models = [
            mid
            for (mid,) in conn.execute(
                sa.select(processing_model.c["id"]).where(
                    processing_model.c.data_generator_id == pdata_id,
                    processing_model.c.id != model_id,
                    processing_model.c.model_type != "zone_proposition",
                )
            )
        ]
        if len(other_models) == 0:
            # Delete model and project data
            # First detach model from project data
            conn.execute(
                project_data.update()
                .values(active_model_id=None, data_id=None)
                .where(project_data.c.id == pdata_id)
            )
            # Delete model
            conn.execute(
                processing_model.delete().where(
                    processing_model.c.id == model_id
                )
            )
            # Remove refs to project data
            conn.execute(
                input_assoc.delete().where(
                    input_assoc.c.project_data_id == pdata_id
                )
            )
            conn.execute(
                input_assoc.delete().where(
                    input_assoc.c.input_data_id == pdata_id
                )
            )
            if data_id is not None:
                conn.execute(
                    sa.text(
                        "DELETE FROM geo_feature "
                        "WHERE id IN ("
                        "SELECT f.id FROM feature f "
                        f"WHERE f.data_id = {data_id})"
                    )
                )
                conn.execute(
                    sa.text(f"DELETE FROM feature WHERE data_id = {data_id}")
                )
                conn.execute(
                    sa.text(
                        "DELETE FROM data_value WHERE attribute_id IN"
                        "(SELECT a.id FROM data_attribute a "
                        f"WHERE a.data_id = {data_id})"
                    )
                )
                conn.execute(
                    sa.text(
                        f"DELETE FROM data_attribute WHERE data_id = {data_id}"
                    )
                )
                conn.execute(
                    base_data.delete().where(base_data.c.id == data_id)
                )
            # Delete project data
            conn.execute(
                project_data.delete().where(project_data.c.id == pdata_id)
            )
        else:
            # Replace active model with other model
            conn.execute(
                project_data.update()
                .values(active_model_id=other_models[0], modified_at=now)
                .where(project_data.c.id == pdata_id)
            )
        count += 1
        progress_bar(count, nb)

    # Remove all inactive zone proposition (all remainings are inactive now)
    op.execute(
        processing_model.delete().where(
            processing_model.c.model_type == "zone_proposition"
        )
    )

    with op.batch_alter_table("processing_model", schema=None) as batch_op:
        batch_op.drop_column("filter_clusters")
        batch_op.drop_column("children")
        batch_op.drop_column("mutation")
        batch_op.drop_column("size")
        batch_op.drop_column("duration")
        batch_op.drop_column("iterations")
        batch_op.drop_column("geo_size")

    # Replace new modeltype with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_modeltype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE processing_model ALTER COLUMN model_type "
        "TYPE _modeltype USING model_type::text::_modeltype"
    )
    new_modeltype.drop(op.get_bind(), checkfirst=False)
    old_modeltype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE processing_model ALTER COLUMN model_type "
        "TYPE modeltype USING model_type::text::modeltype"
    )
    tmp_modeltype.drop(op.get_bind(), checkfirst=False)

    # Replace new geodatatype with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE _geodatatype"
        " USING type::text::_geodatatype"
    )
    new_geodatatype.drop(op.get_bind(), checkfirst=False)
    old_geodatatype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE base_data ALTER COLUMN type TYPE geodatatype"
        " USING type::text::geodatatype"
    )
    tmp_geodatatype.drop(op.get_bind(), checkfirst=False)

    # Replace new task type with old one
    tmp_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE _projecttasktype"
        " USING type::text::_projecttasktype"
    )
    new_task_type.drop(op.get_bind(), checkfirst=False)
    old_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE projecttasktype"
        " USING type::text::projecttasktype"
    )
    tmp_task_type.drop(op.get_bind(), checkfirst=False)



from copy import deepcopy

from . import db
from .geo_data import DataAttribute
from .processing_models import (
    ContinuousRule,
    DiscreteRules,
    MRSort,
    MRSortCriterion,
    PrefDefaultValues,
    WeightedSum,
    _replace_variables_code,
)
from .project import Project
from .project_data import DataGenerator, ProjectData


class ProjectDataProcessingService:
    def rename_data(self, data: ProjectData, new_name: str):
        """Rename project data and modify models using it.

        :param data:
        :param new_name: new project data name
        """
        if data.data is None:
            data.name = new_name
            return

        var_mapping = {}
        for attr in data.data.attributes:
            var_mapping[(data.name, attr.name)] = (new_name, attr.name)

        for d in data.output_data:
            for model in d._processing_models:
                match model:
                    case DiscreteRules():
                        for category in model.categories:
                            rules = []
                            for rule in category.rules:
                                rules.append(
                                    _replace_variables_code(rule, var_mapping)
                                )
                            category.rules = rules
                            db.session.add(category)
                    case ContinuousRule():
                        model.rule = _replace_variables_code(
                            model.rule, var_mapping
                        )
                        db.session.add(model)
        data.name = new_name

    def detach_input(
        self, data: ProjectData, removable_input_data: list[ProjectData]
    ):
        """Detach input from project data.

        Only necessary to remove operands/criteria/default values from
        processing models that used to use those inputs.

        :param data:
        :param removable_input_data: sequence of data inputs to remove

        .. note:: do not actually remove input, simply clean afterwards
        """
        if not isinstance(data, DataGenerator):
            return

        attributes = []
        for d in removable_input_data:
            if d.data is None:
                continue
            attributes += list(d.data.attributes)

        if len(attributes) == 0:
            return

        for model in data._processing_models:
            match model:
                case PrefDefaultValues():
                    for def_value in model.default_values:
                        if def_value.attribute in attributes:
                            db.session.delete(def_value)
            match model:
                case WeightedSum():
                    for op in model.operands:
                        if op.attribute in attributes:
                            db.session.delete(op)
                case MRSort():
                    for criterion in model.criteria:
                        if criterion.attribute in attributes:
                            db.session.delete(criterion)

    def _topological_sort(self, graph):
        """Sort nodes topologically using Kahn's algorithm.

        :param graph: directed graph
        :raises TypeError: if `graph` is not acyclic
        :return: topologically ordered nodes
        """
        nodes = deepcopy(graph)
        sorted_data = []
        set_queue = []
        for node, edges in list(nodes.items()):
            if len(edges) == 0:
                del nodes[node]
                set_queue.append(node)
        while len(set_queue) > 0:
            current = set_queue.pop()
            sorted_data.append(current)
            for node, edges in list(nodes.items()):
                if current in edges:
                    edges.remove(current)
                    if len(edges) == 0:
                        del nodes[node]
                        set_queue.append(node)
        if len(nodes) > 0:
            raise TypeError("Graph is not acyclic")
        return sorted_data

    def get_project_data_graph(
        self, project: Project
    ) -> dict[ProjectData, list[ProjectData]]:
        """Return complete project data graph.

        :param project:
        :return:
        """
        return {data: list(data.input_data) for data in project.data_list}

    def get_input_graph(
        self, data: ProjectData
    ) -> dict[ProjectData, list[ProjectData]]:
        """Return graph of input data spanning from current project data.

        :param data:
        :return:
        """
        graph = {}
        nodes = [data]
        while len(nodes) > 0:
            node = nodes.pop(0)
            if node in graph:
                continue
            graph[node] = list(node.input_data)
            nodes += list(set(node.input_data) - set(nodes))
        return graph

    def get_output_graph(
        self, data: ProjectData
    ) -> dict[ProjectData, list[ProjectData]]:
        """Return graph of output data spanning from current project data.

        :param data:
        :return:
        """
        graph = {}
        nodes = [data]
        while len(nodes) > 0:
            node = nodes.pop(0)
            if node in graph:
                continue
            graph[node] = list(node.output_data)
            nodes += list(set(node.output_data) - set(nodes))
        return graph

    def _replace_attributes_code(
        self,
        data: ProjectData,
        code: str,
        attributes_mapping: dict[DataAttribute, DataAttribute],
    ) -> str:
        """Replace attributes by others in code safely.

        :param data: project data to modify model's code
        :param code: code to modify
        :param attributes_mapping: mapping of old attributes to new ones
        :return: modified code
        """
        var_mapping = {}
        for d in data.project.data_list:
            if d.data is None:
                continue
            for attr in d.data.attributes:
                new_attr = attributes_mapping.get(attr, attr)
                var_mapping[(d.name, attr.name)] = (d.name, new_attr.name)
        return _replace_variables_code(code, var_mapping)

    def replace_input_attributes(
        self,
        data: ProjectData,
        attributes_mapping: dict[DataAttribute, DataAttribute],
    ):
        """Replace input attributes of project data by new ones.

        :param data:
        :param attributes_mapping: mapping of old attributes to new ones
        :raise ValueError:
            * if some new attributes introduced were already in the data inputs
            * if attributes from a different project are used
        :raise TypeError: if an attribute has changed its type
        """
        if not isinstance(data, DataGenerator):
            return

        # Mapping between data id and project data
        project_data_map = {}
        for p_data in data.project.data_list:
            if p_data.data is None:
                continue
            project_data_map[p_data.data.id] = p_data

        # List of data in downstream
        # (that would create cycles if used as inputs)
        unusable_data = sorted(
            set.union(
                *[
                    set(outputs)
                    for outputs in self.get_output_graph(data).values()
                ]
            ),
            key=lambda pdata: pdata.id,
        )
        unusable_data.append(data)  # Also add project data (avoid self loop)

        # Loop over attributes mapping
        new_project_data = set()
        for old_attr, new_attr in attributes_mapping.items():
            # Ignore mapping if old attribute's data is not an input
            if project_data_map[old_attr.data.id] not in data.input_data:
                continue
            # Check new attribute is not already an used input of project data
            if new_attr in data.get_used_input_attributes():
                raise ValueError(
                    "new attribute given was already among used inputs: "
                    f"{new_attr.id}"
                )
            if old_attr.type != new_attr.type:
                raise TypeError(
                    "cannot change attribute type through remapping: "
                    f"{old_attr.id}"
                )
            # Gather set of new project data
            new_project_data.add(project_data_map[new_attr.data.id])

        for new_input_data in new_project_data - set(data.input_data):
            data.add_input(new_input_data)

        for model in data._processing_models:
            match model:
                case PrefDefaultValues():
                    for def_value in model.default_values:
                        if def_value.attribute not in attributes_mapping:
                            continue
                        def_value.attribute = attributes_mapping.get(
                            def_value.attribute, def_value.attribute
                        )
                        db.session.add(def_value)
            match model:
                case WeightedSum():
                    for op in model.operands:
                        if op.attribute not in attributes_mapping:
                            continue
                        op.attribute = attributes_mapping.get(
                            op.attribute, op.attribute
                        )
                        db.session.add(op)
                case MRSort():
                    for criterion in model.criteria:
                        if criterion.attribute not in attributes_mapping:
                            continue
                        new_criterion = MRSortCriterion(
                            mrsort_id=model.id,
                            profiles=criterion.profiles,
                            weight=criterion.weight,
                            maximize=criterion.maximize,
                            attribute_id=(
                                attributes_mapping[criterion.attribute].id,
                            ),
                        )
                        db.session.add(new_criterion)
                        db.session.delete(criterion)
                case DiscreteRules():
                    for category in model.categories:
                        rules = []
                        for rule in category.rules:
                            rules.append(
                                self._replace_attributes_code(
                                    data, rule, attributes_mapping
                                )
                            )
                        category.rules = rules
                        db.session.add(category)
                case ContinuousRule():
                    model.rule = self._replace_attributes_code(
                        data, model.rule, attributes_mapping
                    )
            db.session.add(model)
        data.on_modification()
        db.session.add(data)



"""Replace preference with generic processing

Revision ID: 32025d48c84e
Revises: d1a3e922d5d0
Create Date: 2025-02-14 16:18:30.860615

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql  # noqa: F401

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "32025d48c84e"
down_revision = "d1a3e922d5d0"
branch_labels = None
depends_on = None


preftype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    name="preftype",
)


modeltype = sa.Enum(
    "categories_rule",
    "continuous_rule",
    "geo_buffer",
    "mrsort",
    "weighted_sum",
    "merge_overlap",
    "dissolve_adjacent",
    name="modeltype",
)


feature = sa.Table(
    "feature",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("feature", "geo_feature", name="featuretype"),
        nullable=True,
    ),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column(
        "explainability",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
)


def upgrade():
    op.rename_table("preference_model", "processing_model")
    op.alter_column(
        "processing_model", "pref_type", new_column_name="model_type"
    )
    modeltype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE processing_model ALTER COLUMN model_type TYPE modeltype"
        " USING model_type::text::modeltype"
    )
    preftype.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("default_value", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_default_value_model_id_preference_model", type_="foreignkey"
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_default_value_model_id_processing_model"),
            "processing_model",
            ["model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("discrete_category", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_discrete_category_preference_model_id_preference_model",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_discrete_category_preference_model_id_processing_model"
            ),
            "processing_model",
            ["preference_model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("mrsort_criterion", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_mrsort_criterion_mrsort_id_preference_model",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_mrsort_criterion_mrsort_id_processing_model"),
            "processing_model",
            ["mrsort_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "mrsort_inference_alternative", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_mrsort_inference_alternative_mrsort_id_preference_model",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_mrsort_inference_alternative_mrsort_id_processing_model"
            ),
            "processing_model",
            ["mrsort_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_project_data_active_model_id_preference_model",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_project_data_active_model_id_processing_model"),
            "processing_model",
            ["active_model_id"],
            ["id"],
        )

    with op.batch_alter_table("weighted_sum_operand", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_weighted_sum_operand_model_id_preference_model",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_weighted_sum_operand_model_id_processing_model"),
            "processing_model",
            ["model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    conn = op.get_bind()
    # Get all features with an explanation
    nb = conn.execute(
        sa.text("SELECT count(id) FROM feature WHERE explainability != '[]'")
    ).fetchone()[0]
    res = conn.execute(
        sa.select(feature.c["id", "explainability"]).where(
            feature.c.explainability != "[]"
        )
    )
    i = 0
    progress_bar(i, nb)
    for id, explain in res:
        explain["model_type"] = explain.pop("pref_type")
        conn.execute(
            feature.update()
            .values(explainability=explain)
            .where(feature.c.id == id)
        )
        i += 1
        progress_bar(i, nb)


def downgrade():
    op.rename_table("processing_model", "preference_model")
    op.alter_column(
        "preference_model", "model_type", new_column_name="pref_type"
    )
    preftype.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE preference_model ALTER COLUMN pref_type TYPE preftype"
        " USING pref_type::text::preftype"
    )
    modeltype.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("weighted_sum_operand", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_weighted_sum_operand_model_id_processing_model"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_weighted_sum_operand_model_id_preference_model",
            "preference_model",
            ["model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_project_data_active_model_id_processing_model"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_project_data_active_model_id_preference_model",
            "preference_model",
            ["active_model_id"],
            ["id"],
        )

    with op.batch_alter_table(
        "mrsort_inference_alternative", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "fk_mrsort_inference_alternative_mrsort_id_processing_model"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_mrsort_inference_alternative_mrsort_id_preference_model",
            "preference_model",
            ["mrsort_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("mrsort_criterion", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_mrsort_criterion_mrsort_id_processing_model"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_mrsort_criterion_mrsort_id_preference_model",
            "preference_model",
            ["mrsort_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("discrete_category", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "fk_discrete_category_preference_model_id_processing_model"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_discrete_category_preference_model_id_preference_model",
            "preference_model",
            ["preference_model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("default_value", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_default_value_model_id_processing_model"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_default_value_model_id_preference_model",
            "preference_model",
            ["model_id"],
            ["id"],
            ondelete="CASCADE",
        )

    conn = op.get_bind()
    # Get all features with an explanation
    nb = conn.execute(
        sa.text("SELECT count(id) FROM feature WHERE explainability != '[]'")
    ).fetchone()[0]
    res = conn.execute(
        sa.select(feature.c["id", "explainability"]).where(
            feature.c.explainability != "[]"
        )
    )
    i = 0
    progress_bar(i, nb)
    for id, explain in res:
        explain["pref_type"] = explain.pop("model_type")
        conn.execute(
            feature.update()
            .values(explainability=explain)
            .where(feature.c.id == id)
        )
        i += 1
        progress_bar(i, nb)



"""This module gathers functions/classes to extract openAPI specification
from this API.

It can be used as a python function or script.

.. code:: python

    from deseasion.backend.openapi.extract_specification import extract_openapi

    spec = extract_openapi()

.. code:: bash

    python -m deseasion.backend.openapi.extract_specification
"""

import importlib.resources
import inspect
import logging
import re
import sys
from typing import Type

import yaml
from apispec import APISpec, BasePlugin
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.common import (
    get_unique_schema_name,
    make_schema_key,
    resolve_schema_instance,
)
from flask_restful import Api, Resource
from marshmallow import Schema, fields

from ..resources import api
from ..resources.utils import type2openapi
from ..schemas import *  # noqa: F403 (import all schemas for processing)


def get_schemas_classes(
    namespace: str | None = None,
) -> dict[str, Type[BaseSchema]]:  # noqa: F405
    """Get all schema classes from :mod:`api`.

    :param namespace:
        namespace from which to extract specification
        Defaults: ``__name__``
    :return: schema classes as a dictionary
    """
    namespace = namespace or __name__
    res = {}
    for name, obj in inspect.getmembers(sys.modules[namespace]):
        if inspect.isclass(obj) and issubclass(obj, Schema) and obj != Schema:
            res[name.replace("Schema", "")] = obj
    return res


def enum_fields2properties(self, field, **kwargs):
    """Helper function to convert enum field to openAPI spec.

    :param field:
    :return: openAPI spec
    """
    res = {}
    if isinstance(field, fields.Enum):
        values = []
        for member in field.enum:
            values.append(member.name)
        res["type"] = "string"
        res["enum"] = values
    return res


def pluck_fields2properties(self, field, **kwargs) -> dict:
    """Helper function to convert pluck field to openAPI spec.

    It uses a reference on the plucked field instead of its type.

    :param field:
    :return: openAPI spec
    """
    res = {}
    if isinstance(field, fields.Pluck):
        nested = (
            field.nested
            if isinstance(field.nested, str)
            else field.nested.__name__
        )
        ref = nested.replace("Schema", "")
        ref_field = f"#/components/schemas/{ref}/properties/{field.only[0]}"
        if field.many:
            res["items"] = res.get("items", {})
            res["items"]["$ref"] = ref_field
        else:
            res["$ref"] = ref_field
    return res


def method_fields2properties(self, field, **kwargs) -> dict:
    """Helper function to convert method field to openAPI spec.

    It uses the `field.field` field as the actual field to convert.

    :param field:
    :return: openAPI spec
    """
    ret = {}
    if isinstance(field, FieldMethod):  # noqa: F405
        ret = self.field2property(field.field)
    return ret


class MyMarshmallowPlugin(MarshmallowPlugin):
    """This class is a modified version of the basic Marshmallow plugin.

    It is capable of handling :class:`api.schemas.OneOfSchema` schemas.
    """

    schema_names = {}

    def schema_helper(self, name, _, schema=None, **kwargs):
        """Definition helper that allows using a marshmallow
        :class:`Schema <marshmallow.Schema>` to provide OpenAPI
        metadata.

        It was modified to handle correctly :class:`api.schemas.OneOfSchema`
        schemas.

        :param type|Schema schema: A marshmallow Schema class or instance.
        """
        if schema is None:
            return None

        schema_instance = resolve_schema_instance(schema)

        schema_key = make_schema_key(schema_instance)
        self.warn_if_schema_already_in_spec(schema_key)
        assert self.converter is not None, "init_spec has not yet been called"
        self.converter.refs[schema_key] = name

        if isinstance(schema_instance, OneOfSchema):  # noqa: F405
            json_schema = {
                "oneOf": [],
                "discriminator": {
                    "propertyName": schema_instance.type_field,
                    "mapping": {},
                },
            }
            for subname, sch_init in schema_instance.type_schemas.items():
                sch = sch_init(openapi=True)
                subsch_key = make_schema_key(sch)
                if subsch_key not in self.converter.refs:
                    new_subname = self.schema_names[sch.__class__]
                    self.spec.components.schema(
                        get_unique_schema_name(
                            self.spec.components, new_subname
                        ),
                        schema=sch,
                    )
                sch_ref = (
                    f"#/components/schemas/{self.converter.refs[subsch_key]}"
                )
                json_schema["oneOf"].append({"$ref": sch_ref})
                json_schema["discriminator"]["mapping"][subname] = sch_ref
            return json_schema

        return self.converter.schema2jsonschema(schema_instance)


def deep_update(dico: dict, new_dico: dict):
    """Perform a deep update on the first dictionnary with content of second.

    It handles nested dictionary structures, and lists.
    Everything else is updated the regular way.

    :param dico: dictionnary to update
    :param new_dico: updated data as a dictionnary
    """
    for k, v in new_dico.items():
        if k not in dico:
            dico[k] = v
        elif (  # noqa: E721
            type(dico[k]) == list and type(v) == list
        ):  # Make sure to exclude strings
            dico[k] += v
        elif isinstance(dico[k], dict) and isinstance(v, dict):
            deep_update(dico[k], v)
        else:
            dico[k] = v


class FlaskRestfulPlugin(BasePlugin):
    """This class is a :mod:`apispec` plugin made to handle
    :class:`flask_restful.Resource`.
    """

    def init_spec(self, spec: APISpec) -> None:
        """Initialize plugin with APISpec object

        :param APISpec spec: APISpec object this plugin instance is attached to
        """
        self.spec = spec

    def path_helper(
        self,
        path: str | None = None,
        operations: dict | None = None,
        parameters: list[dict] | None = None,
        *,
        resource: Type[Resource] | None = None,
        api: Api | None,
    ):
        """Add path to openAPI specification.

        :param api: ref to API
        :param path: endpoint of path (unused here)
        :param operations: operations statically defined on path
        :param parameters: parameters statically defined on path
        :param resource: API resource
        :raises ValueError: if `resource` is not found in `api`
        :return: openAPI path specification
        """
        assert resource is not None
        assert api is not None

        parameters = [] or parameters
        operations = {} or operations

        for klass, endpoints, _ in api.resources:
            if klass == resource:
                res = self.resource2openapi(klass, endpoints[0])
                operations.update(res["operations"])
                parameters += res.get("parameters", [])
                return res["path"]
        raise ValueError(f"Resource {resource.__name__} not found in API")

    @staticmethod
    def extract_doc_strings(func):
        """Extract any openAPI useful information from operation doc-string.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI specification
        """
        res = {}
        doc = func.__doc__
        if doc is None:
            return res
        res["description"] = ""
        lines = iter(doc.splitlines())
        # Read description
        while (line := next(lines, None)) is not None:
            line = line.lstrip()
            if line.startswith(":"):
                break
            else:
                res["description"] += "\n" + line
        res["description"] = res["description"].strip()
        # Read rest of doc string
        tree = []
        current_node = None
        while line is not None:
            line = line.strip()
            if line == "---":
                # Reaching yaml part => abort
                break
            if re.match(r"^:", line):
                header = re.findall(r"^:[^:]*:", line)[0]
                if header.startswith(":param"):
                    # Parsing a path parameter
                    description = line[len(header) :].strip()
                    type_, arg = (
                        re.sub(r"^:param ", "", header)
                        .rstrip()[:-1]
                        .split(" ")
                    )
                    current_node = ["param", arg, type_, description]
                elif header.startswith(":status"):
                    # Parsing a response status
                    description = line[len(header) :].strip()
                    status = re.findall(r"\d+", header)[0]
                    current_node = ["status", status, description]
                else:
                    current_node = None
                tree.append(current_node)
            elif re.match(r"^.. :quickref: ", line):
                header = re.findall(r"^.. :quickref: ", line)[0]
                quickref = line[len(header) :]
                tag, description = quickref.split(";")[:2]
                current_node = ["quickref", tag, description]
                tree.append(current_node)
            elif current_node is not None:
                current_node[-1] += " " + line
            line = next(lines, None)

        for node in tree:
            match node:
                case None:
                    continue
                case ["param", arg, type_, description]:
                    res["parameters"] = res.get("parameters", [])
                    res["parameters"].append(
                        {
                            "name": arg,
                            "in": "path",
                            "required": True,
                            "schema": {"type": type2openapi(type_)},
                            "description": description.rstrip().lstrip(),
                        }
                    )
                case ["status", status, description]:
                    res["responses"] = res.get("responses", {})
                    res["responses"][str(status)] = {
                        "description": description.lstrip().rstrip()
                    }
                case ["quickref", tag, description]:
                    res["tags"] = res.get("tags", [])
                    res["tags"].append(tag)
                    res["summary"] = description.rstrip().lstrip()
        return res

    def resource2openapi(self, resource: Type[Resource], endpoint: str):
        """Convert API resource to openAPI specication.

        It extract spec from doc-string and special decorators
        (those modifying `func.__apispec__`).

        :param resource:
        :param endpoint: path endpoint
        :return: _description_
        """
        res = {"path": endpoint, "operations": {}}
        for method in resource.methods:
            func = getattr(resource, method.lower())
            operation = FlaskRestfulPlugin.extract_doc_strings(func)
            deep_update(operation, self.apispec_attr2openapi(func) or {})
            operation["responses"] = operation.get("responses", {})
            if "200" not in operation["responses"]:
                operation["responses"]["200"] = {"description": ""}
            res["operations"][method.lower()] = operation

        groups = re.findall(r"<[^<]*>", endpoint)
        if len(groups) == 0:
            return res
        res["parameters"] = []
        for group in groups:
            type_, arg = re.sub(r"[<>]", "", group).split(":")
            res["parameters"].append(
                {
                    "name": arg,
                    "in": "path",
                    "required": True,
                    "schema": {"type": type2openapi(type_)},
                }
            )
            res["path"] = res["path"].replace(group, f"{{{arg}}}")
        return res

    @property
    def marshmallow_plugin(self) -> MarshmallowPlugin | None:
        """Return :class:`MarshmallowPlugin` plugin if it is used to build
        the specification.

        :return:
        """
        for plugin in self.spec.plugins:
            if isinstance(plugin, MarshmallowPlugin):
                return plugin
        return None

    def _find_schema(self, schema: Schema) -> str | None:
        """Find name under which a schema is registered.

        :param schema:
        :return: registered name or ``None``
        """
        key = make_schema_key(schema)
        return self.marshmallow_plugin.converter.refs.get(key, None)

    def apispec_attr2openapi(self, func) -> dict:
        """Extract operation openAPI spec from function decorators.

        It extract spec from special decorators
        (those modifying `func.__apispec__`).

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        return {
            **self.apispec_parameters2openapi(func),
            **self.apispec_responses2openapi(func),
            **self.apispec_requestBody2openapi(func),
            **self.apispec_security2openapi(func),
        }

    def apispec_security2openapi(self, func) -> dict:
        """Extract operation security openAPI spec from function decorators.

        It extract spec from `func.__apispec__['security']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "security" not in apispec_attr:
            return {}
        return {"security": apispec_attr.get("security", [])}

    def apispec_parameters2openapi(self, func) -> dict:
        """Extract operation parameters openAPI spec from function decorators.

        It extract spec from `func.__apispec__['parameters']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "parameters" not in apispec_attr:
            return {}
        return {"parameters": apispec_attr.get("parameters", [])}

    def apispec_requestBody2openapi(self, func) -> dict:
        """Extract operation request body openAPI spec from function
        decorators.

        It extract spec from `func.__apispec__['requestBody']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "requestBody" not in apispec_attr:
            return {}
        request = apispec_attr.get("requestBody", {})

        res = {}
        res["requestBody"] = {
            "description": request.get("description", ""),
            "content": request.get("content", {}),
            "required": request.get("required", True),
        }
        schema = request.get("schema", None)
        if isinstance(schema, str):
            deep_update(
                res["requestBody"]["content"],
                {"application/json": {"schema": {"$ref": schema}}},
            )
        elif isinstance(schema, list):
            names = []
            if self.marshmallow_plugin:
                for sch in schema:
                    schema_instance = resolve_schema_instance(sch)
                    name = self._find_schema(schema_instance)
                    if name is None:
                        name = get_unique_schema_name(
                            self.spec.components,
                            type(schema_instance).__name__,
                        )
                        self.spec.components.schema(
                            name, schema=schema_instance
                        )
                    names.append(name)
            deep_update(
                res["requestBody"]["content"],
                {
                    "application/json": {
                        "schema": {
                            "oneOf": [
                                {"$ref": f"#/components/schemas/{name}"}
                                for name in names
                            ]
                        }
                    }
                },
            )
        elif isinstance(schema, type) or isinstance(schema, Schema):
            schema_instance = resolve_schema_instance(schema)
            if self.marshmallow_plugin:
                name = self._find_schema(schema_instance)
                if name is None:
                    name = get_unique_schema_name(
                        self.spec.components, type(schema_instance).__name__
                    )
                    self.spec.components.schema(name, schema=schema_instance)
                if isinstance(schema, Schema) and schema.many:
                    deep_update(
                        res["requestBody"]["content"],
                        {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {
                                        "$ref": f"#/components/schemas/{name}"
                                    },
                                }
                            }
                        },
                    )
                else:
                    deep_update(
                        res["requestBody"]["content"],
                        {
                            "application/json": {
                                "schema": {
                                    "$ref": f"#/components/schemas/{name}"
                                }
                            }
                        },
                    )
        return res

    def apispec_responses2openapi(self, func) -> dict:
        """Extract operation responses openAPI spec from function decorators.

        It extract spec from `func.__apispec__['responses']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "responses" not in apispec_attr:
            return {}

        res = {}
        res["responses"] = {}
        for status, response in apispec_attr.get("responses", {}).items():
            res["responses"][str(status)] = {
                "description": response.get("description", ""),
                "content": response.get("content", {}),
            }
            schema = response.get("schema", None)
            if isinstance(schema, str):
                deep_update(
                    res["responses"][str(status)]["content"],
                    {"application/json": {"schema": {"$ref": schema}}},
                )
            elif isinstance(schema, list):
                names = []
                if self.marshmallow_plugin:
                    for sch in schema:
                        schema_instance = resolve_schema_instance(sch)
                        name = self._find_schema(schema_instance)
                        if name is None:
                            name = get_unique_schema_name(
                                self.spec.components,
                                type(schema_instance).__name__,
                            )
                            self.spec.components.schema(
                                name, schema=schema_instance
                            )
                        names.append(name)
                deep_update(
                    res["responses"][str(status)]["content"],
                    {
                        "application/json": {
                            "schema": {
                                "oneOf": [
                                    {"$ref": f"#/components/schemas/{name}"}
                                    for name in names
                                ]
                            }
                        }
                    },
                )
            elif isinstance(schema, type) or isinstance(schema, Schema):
                schema_instance = resolve_schema_instance(schema)
                if self.marshmallow_plugin:
                    name = self._find_schema(schema_instance)
                    if name is None:
                        name = get_unique_schema_name(
                            self.spec.components,
                            type(schema_instance).__name__,
                        )
                        self.spec.components.schema(
                            name, schema=schema_instance
                        )
                    if isinstance(schema, Schema) and schema.many:
                        deep_update(
                            res["responses"][str(status)]["content"],
                            {
                                "application/json": {
                                    "schema": {
                                        "type": "array",
                                        "items": {
                                            "$ref": (
                                                f"#/components/schemas/{name}"
                                            )
                                        },
                                    }
                                }
                            },
                        )
                    else:
                        deep_update(
                            res["responses"][str(status)]["content"],
                            {
                                "application/json": {
                                    "schema": {
                                        "$ref": f"#/components/schemas/{name}"
                                    }
                                }
                            },
                        )

        return res


def extract_openapi(namespace: str | None = None) -> APISpec:
    logging.info("Extract openAPI specification")
    spec_path = importlib.resources.files(
        "deseasion.backend.openapi"
    ).joinpath("spec_geo_json.yml")
    with spec_path.open("r") as f:
        geo_json_spec = yaml.safe_load(f)

    ma_plugin = MyMarshmallowPlugin()
    plugins = [ma_plugin, FlaskRestfulPlugin()]
    spec = APISpec(
        "DESEASION API",
        "2.0.4",
        "3.0.2",
        servers=[
            {
                "url": "/api",
                "description": "Production Server (behind reverse-proxy)",
            },
            {
                "url": "/",
                "description": (
                    "Test Server (backend only without reverse proxy)"
                ),
            },
        ],
        components=geo_json_spec.get("components", {}),
        plugins=plugins,
    )
    ma_plugin.converter.add_attribute_function(enum_fields2properties)
    ma_plugin.converter.add_attribute_function(method_fields2properties)

    schema_classes = get_schemas_classes(namespace)
    ma_plugin.schema_names = {klass: k for k, klass in schema_classes.items()}

    for k, klass in schema_classes.items():
        schema_instance = resolve_schema_instance(klass)
        schema_key = make_schema_key(schema_instance)
        if schema_key not in ma_plugin.converter.refs:
            spec.components.schema(
                get_unique_schema_name(spec.components, k), schema=klass
            )
    logging.debug(
        f"Schemas added:{len(spec.components.schemas)} found: "
        f"{len(schema_classes)}"
    )

    spec.components.security_scheme(
        "bearerAuth",
        {"type": "http", "scheme": "bearer", "bearerFormat": "JWT"},
    )
    spec.components.security_scheme(
        "basicAuth",
        {
            "type": "http",
            "scheme": "basic",
            "description": "Provide access token",
        },
    )
    spec.components.security_scheme(
        "refreshAuth",
        {
            "type": "http",
            "scheme": "bearer",
            "bearerFormat": "JWT",
            "description": "Provide refresh token",
        },
    )

    for klass, _, _ in api.resources:
        spec.path(resource=klass, api=api)

    return spec


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    spec = extract_openapi()
    print(spec.to_yaml())



"""This module implements requests functions to external servers.
"""

import requests
from flask import current_app as app

from ..exceptions import ExternalRequestError


def get_external_timeout() -> float:
    """Return external requests timeout.

    Taken from app config 'EXTERNAL_REQUEST_TIMEOUT', defaults to 10.

    :return:
    """
    return app.config.get("EXTERNAL_REQUEST_TIMEOUT", 10)


def get_requests_proxies() -> dict | None:
    """Return set up requests proxies.

    Taken from app config 'HTTP_PROXY' and 'HTTPS_PROXY'.
    Defaults to `None` if no proxy set up.

    :return:
    """
    proxies = {}
    if "HTTP_PROXY" in app.config:
        proxies["http"] = app.config["HTTP_PROXY"]
    if "HTTPS_PROXY" in app.config:
        proxies["https"] = app.config["HTTPS_PROXY"]
    if len(proxies) > 0:
        return proxies
    return None


def get(
    url: str,
    params: dict | None,
    timeout: float | None = None,
    request_type: str | None = None,
    proxies: dict | None = None,
    **kwargs,
) -> requests.Response:
    """Perform GET request and return response.

    :param url:
    :param params: query arguments
    :param timeout: request timeout
    :param request_type: request type (only used in raised error if this fails)
    :param proxies: requests proxy
    :raises ExternalRequestError: if request fails
    :return:
    """
    timeout = timeout or get_external_timeout()
    proxies = proxies or get_requests_proxies()
    try:
        response = requests.get(
            url, params=params, timeout=timeout, proxies=proxies, **kwargs
        )
        response.raise_for_status()
        return response
    except requests.Timeout as e:
        raise ExternalRequestError(
            e.request.url, "request timeout", request_type
        )
    except requests.ConnectionError as e:
        raise ExternalRequestError(
            e.request.url, "could not reach server", request_type
        )
    except requests.HTTPError as e:
        raise ExternalRequestError(
            e.request.url,
            f"server responded with error {e.response.status_code}",
            request_type,
        )
    except requests.RequestException as e:
        if e.request is None:
            raise ExternalRequestError(
                url, "Server URL is invalid", request_type
            )
        raise ExternalRequestError(e.request.url, str(e), request_type)



from flask import jsonify, request
from flask_restful import Resource, reqparse
from sqlalchemy.exc import IntegrityError

from ..exceptions import RequestError
from ..models import DataAttribute, DataType, PermissionAbility, Project, db
from ..models.project_data_utils import ProjectDataProcessingService
from ..schemas import (
    DataGeneratorSchema,
    DataGeoSchema,
    DataStreamSchema,
    MessageSchema,
    ProjectAccessSchema,
    ProjectCreationSchema,
    ProjectDataBaseSchema,
    ProjectDataCreationSchema,
    ProjectDataReplaceInputsPostRequestBodySchema,
    ProjectFromTemplateRequestBody,
    ProjectGetResponseSchema,
    ProjectGlobalDataSchema,
    ProjectListGetResponseSchema,
    ProjectPermissionGetResponseSchema,
    ProjectSchema,
    ProjectSharedDataListGetResponseSchema,
    ProjectSharedDataListPostResponseSchema,
    ProjectTemplateListGetResponseSchema,
    ProjectTemplatePostResponseSchema,
    TemplateAccessSchema,
    TemplatePermissionGetResponseSchema,
    TemplateSchema,
)
from ..services import (
    geo_data_service,
    geo_data_stream_service,
    project_data_service,
    project_service,
)
from ..services.auth_service import check_jwt_authentication, token_required
from ..services.permission_service import has_ability
from .utils import (
    dump_data,
    get_json_content,
    with_query_arg,
    with_request_body,
    with_response,
)


class ProjectListAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectListGetResponseSchema,
        description="Full list of accessible projects",
    )
    def get(self):
        """
        Return the full list of projects accessible by the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).

        .. :quickref: Project; Get the list of accessible projects
        """
        projects = project_service.get_all_authorized_projects()
        return dump_data(ProjectSchema(many=True), projects=projects)

    @token_required
    @has_ability(PermissionAbility.create_project)
    @with_response(
        200,
        schema=ProjectGetResponseSchema,
        description="Created project details",
    )
    @with_request_body(
        schema=[ProjectCreationSchema, ProjectFromTemplateRequestBody],
        description=(
            "Either the project details or the template from which to create "
            "it"
        ),
    )
    def post(self):
        """
        Create a new project.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to create projects.

        .. :quickref: Project; Create a new project
        """
        user = check_jwt_authentication(request)
        content = get_json_content()
        schema = ProjectCreationSchema()
        if "template" in content.keys():
            # load the project from the template using the template id
            template_id = content.pop("template")
            with db.session.no_autoflush:
                project = project_service.create_project_from_template(
                    template_id, user
                )
            print(project, project.is_template)
            # update the project with the content
            project = schema.load(content, instance=project)
        else:
            project = schema.load(content)
            project.manager = user
        project.create()
        return dump_data(schema, project=project)


class ProjectAPI(Resource):
    @token_required
    @with_response(
        200, schema=ProjectGetResponseSchema, description="Project details"
    )
    def get(self, project_id):
        """
        Return the project details for a given id.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Get the project details
        """
        project = project_service.get_if_authorized(project_id)
        return dump_data(ProjectSchema(), project=project)

    @token_required
    @with_response(
        200, schema=ProjectGetResponseSchema, description="New project details"
    )
    @with_request_body(schema=ProjectSchema, description="New project details")
    def put(self, project_id):
        """
        Update the project details.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to modify this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Update the project details
        """
        project = project_service.get_if_authorized(project_id)
        content = get_json_content()
        schema = ProjectSchema()
        project = schema.load(content, instance=project)
        project.update()
        return dump_data(schema, project=project)

    @token_required
    @with_response(
        200, schema=MessageSchema, description="Project successfully deleted"
    )
    def delete(self, project_id):
        """
        Delete the project.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400:
            The project is referenced by other objects and cannot be deleted.
        :status 403: The user is not allowed to delete this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Delete the project
        """
        project = project_service.get_if_authorized(project_id)
        try:
            project.delete()
        except IntegrityError:
            raise RequestError(
                (
                    "The project is referenced by other objects and cannot be "
                    "deleted"
                ),
                400,
            )
        return jsonify(message="project {} deleted".format(project_id))


class ProjectPermissionsAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectPermissionGetResponseSchema,
        description="Project permissions",
    )
    def get(self, project_id):
        """
        Allow the project manager to get the list of permissions.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the manager of this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Get the project permissions
        """
        user = check_jwt_authentication(request)
        project = project_service.get_if_manager(project_id, user)
        schema = ProjectAccessSchema()
        return dump_data(schema, access=project)

    @token_required
    @with_response(
        200,
        schema=ProjectPermissionGetResponseSchema,
        description="New project permissions",
    )
    @with_request_body(
        schema=ProjectAccessSchema, description="New project permissions"
    )
    def put(self, project_id):
        """
        Allow to modify the list of permissions for the project.

        Only the project manager is allowed to modify the permissions.
        The project manager is always added to the list of authorized users.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the manager of this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Modify the project permissions
        """
        user = check_jwt_authentication(request)
        project = project_service.get_if_manager(project_id, user)
        schema = ProjectAccessSchema()
        content = get_json_content()
        project = schema.load(content, instance=project)
        if not any([p.user is project.manager for p in project.permissions]):
            # always authorize the manager of the project
            project.permissions.append(
                Project.Permission(user=project.manager)
            )
        project.update()
        return dump_data(schema, access=project)


class ProjectSharedDataListAPI(Resource):
    def load_content(self):
        """Loads the content and the data_type from the request body"""
        content = get_json_content()
        try:
            data_type_str = content.pop("data_type")
            data_type = DataType[data_type_str]
        except KeyError:
            raise RequestError(
                'The json data should contain a "data_type" field with a one '
                "of the following values: "
                "{}".format(", ".join(d.name for d in DataType)),
                400,
            )
        return content, data_type

    @token_required
    @with_response(
        200,
        schema=ProjectSharedDataListPostResponseSchema,
        description="List of project shared data",
    )
    @with_request_body(
        schema=ProjectDataCreationSchema,
        description="New project data details",
    )
    def post(self, project_id):
        """
        Create a new project data in the project.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: There is an error in the json schema for the data,
                     or the data/stream already exists in the project.
        :status 403: The user is not allowed to access this project, the data
                     or the stream.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Create a new data in the project
        """
        project = project_service.get_if_authorized(project_id)
        content, data_type = self.load_content()
        schema = ProjectDataCreationSchema.type_schemas[data_type.name]
        try:
            input_data = content.pop("input_data", [])
            output_data = content.pop("output_data", [])
            project_data = schema().load(content)
        except NameError as error:
            raise RequestError(str(error), 400)
        if project_data.data_type != DataType.generator:
            if project_data.data_type in (
                DataType.geo_data,
                DataType.global_data,
            ):
                geo_data_service.check_permission(project_data.data)
            else:
                geo_data_stream_service.check_permission(project_data.stream)
            for d in project.data_list:
                if project_data.data_type in (
                    DataType.geo_data,
                    DataType.global_data,
                ):
                    if d.data is not None and project_data.data == d.data:
                        raise RequestError(
                            "This data is already shared in the project", 400
                        )
                elif d.data_type == project_data.data_type:
                    # Both are DataStream
                    if (
                        d.stream is not None
                        and project_data.stream == d.stream
                    ):
                        raise RequestError(
                            "This stream is already shared in the project", 400
                        )
        project_data.project = project

        try:
            for data_input_id in input_data:
                project_data.add_input(
                    project_data_service.get_if_authorized(data_input_id)
                )

            for data_output_id in output_data:
                output = project_data_service.get_if_authorized(data_output_id)
                output.add_input(project_data)
                output.on_modification()
                db.session.add(output)
        except (TypeError, ValueError) as exc:
            raise RequestError(str(exc), 400)

        ps = ProjectDataProcessingService()
        data_graph = ps.get_project_data_graph(project)
        # Need to make graph with ids otherwise topological sort will create
        # new ProjectData when cloning the dictionary
        graph = {
            p.id: [idata.id for idata in inputs]
            for p, inputs in data_graph.items()
        }
        try:
            ps._topological_sort(graph)
        except TypeError:
            raise RequestError(
                "new data input creates cycles in project data graph", 400
            )

        project_data.create()
        dump_sch = ProjectDataBaseSchema(
            only=(
                "name",
                "id",
                "data",
                "data_type",
                "input_data",
                "output_data",
            ),
            many=True,
        )
        return dump_data(dump_sch, data_list=project.data_list)

    @token_required
    @with_response(
        200,
        schema=ProjectSharedDataListGetResponseSchema,
        description="List of project shared data",
    )
    def get(self, project_id):
        """
        Return the list of project data in the project.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Get the list of data in the project
        """
        project = project_service.get_if_authorized(project_id)
        dump_schemas = {
            DataType.geo_data: DataGeoSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.generator: DataGeneratorSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "processing_model",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.global_data: ProjectGlobalDataSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.data_stream: DataStreamSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                    "stream",
                )
            ),
        }
        res = []
        for projdata in project.data_list:
            res.append(dump_schemas[projdata.data_type].dump(projdata))
        return {"data_list": res}


class ProjectReplaceInputAttributesAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectSharedDataListPostResponseSchema,
        description="List of project data",
    )
    @with_request_body(schema=ProjectDataReplaceInputsPostRequestBodySchema)
    def post(self, project_id):
        """Replace project data attributes usage.

        It replaces usage of old attributes given by new attributes given,
        assuming the old attributes given are attributes of this project's data
        and that the new attributes given are not already used as inputs
        alongside some of the old attributes.

        Its main usage is threfore to replace whole project data
        (all their attributes) by completely unused other project data.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: Remapping could not proceed with given arguments.
        :status 403: The user is not allowed to access this project, or a data
        :status 404: The project or a data doesn't exist.

        .. :quickref: Project; Replace project data attributes usage
        """
        ps = ProjectDataProcessingService()

        content = get_json_content()
        args = ProjectDataReplaceInputsPostRequestBodySchema().load(content)

        attributes_mapping = {}
        for item in args["attributes_mapping"]:
            old_attr_id = item["old_attribute"]
            new_attr_id = item["new_attribute"]
            old_attr = DataAttribute.get_by_id(old_attr_id)
            new_attr = DataAttribute.get_by_id(new_attr_id)
            if old_attr is None or new_attr is None:
                raise RequestError("attributes mapping is invalid", 400)
            attributes_mapping[old_attr] = new_attr
        try:
            project = project_service.replace_input_attributes(
                project_id, attributes_mapping
            )
        except (ValueError, TypeError) as exc:
            raise RequestError(str(exc), 400)

        data_graph = ps.get_project_data_graph(project)
        # Need to make graph with ids otherwise topological sort will create
        # new ProjectData when cloning the dictionary
        graph = {
            p.id: [idata.id for idata in inputs]
            for p, inputs in data_graph.items()
        }
        try:
            ps._topological_sort(graph)
        except TypeError:
            raise RequestError(
                "new data input creates cycles in project data graph", 400
            )
        db.session.commit()

        dump_sch = ProjectDataBaseSchema(
            only=(
                "name",
                "id",
                "data",
                "data_type",
                "input_data",
                "output_data",
            ),
            many=True,
        )
        return dump_data(dump_sch, data_list=project.data_list)


class ProjectTemplateAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectTemplatePostResponseSchema,
        description="Created template details",
    )
    def post(self, project_id):
        """Create a template from a project.

        :param int project_id: The id of the project.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Create a template from project
        """
        user = check_jwt_authentication(request)
        with db.session.no_autoflush:
            template = project_service.create_template_from_project(
                project_id, user
            )
        schema = TemplateSchema()
        content = get_json_content()
        template = schema.load(content, instance=template)
        template.create()
        return dump_data(schema, template=template)


class TemplateListAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectTemplateListGetResponseSchema,
        description="List of accessible project templates",
    )
    @with_query_arg(
        "managed",
        bool,
        required=False,
        description=(
            "Return all accessible templates if false or absent, "
            "only managed templates otherwise",
        ),
    )
    def get(self):
        """
        Return the list of templates of the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).

        .. :quickref: Template; Get the list of accessible templates
        """
        parser = reqparse.RequestParser()
        parser.add_argument("managed", type=bool, location="args")
        args = parser.parse_args()
        user = check_jwt_authentication(request)
        if args.get("managed", False):
            templates = project_service.get_managed_templates(user)
        else:
            templates = project_service.get_user_templates(user)
        return dump_data(TemplateSchema(many=True), templates=templates)


class TemplateAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectTemplatePostResponseSchema,
        description="New project template details",
    )
    @with_request_body(
        schema=TemplateSchema, description="New project template details"
    )
    def put(self, template_id):
        """Modify a project template.

        :param int template_id: The id of the project template.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the template.
        :status 404: The template doesn't exist.

        .. :quickref: Template; Modify a project template
        """
        user = check_jwt_authentication(request)
        template = project_service.get_template_if_manager(template_id, user)
        content = get_json_content()
        schema = TemplateSchema()
        template = schema.load(content, instance=template)
        template.update()
        return dump_data(schema, template=template)

    @token_required
    @with_response(
        200, schema=MessageSchema, description="Template deleted successfully"
    )
    def delete(self, template_id):
        """Delete a project template.

        :param int template_id: The id of the project template.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to delete the template.
        :status 404: The template doesn't exist.

        .. :quickref: Template; Delete a project template
        """
        user = check_jwt_authentication(request)
        template = project_service.get_template_if_manager(template_id, user)
        template.delete()
        return jsonify(message="template {} deleted".format(template_id))


class TemplatePermissionsAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=TemplatePermissionGetResponseSchema,
        description="Template permissions",
    )
    def get(self, template_id):
        """
        Allow the template manager to get the list of permissions.

        :param int template_id: The id of the template.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the manager of this template.
        :status 404: The template doesn't exist.

        .. :quickref: Template; Get the template permissions
        """
        user = check_jwt_authentication(request)
        template = project_service.get_template_if_manager(template_id, user)
        schema = TemplateAccessSchema()
        return dump_data(schema, access=template)

    @token_required
    @with_response(
        200,
        schema=TemplatePermissionGetResponseSchema,
        description="New template permissions",
    )
    @with_request_body(
        schema=TemplateAccessSchema, description="New template permissions"
    )
    def put(self, template_id):
        """
        Allow to modify the list of permissions for the template.

        Only the template manager is allowed to modify the permissions.
        The template manager is always added to the list of authorized users.

        :param int template_id: The id of the template.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the manager of this template.
        :status 404: The template doesn't exist.

        .. :quickref: Template; Modify the template permissions
        """
        user = check_jwt_authentication(request)
        template = project_service.get_template_if_manager(template_id, user)
        schema = ProjectAccessSchema()
        content = get_json_content()
        template = schema.load(content, instance=template)
        if not any([p.user is template.manager for p in template.permissions]):
            # always authorize the manager of the project
            template.permissions.append(
                Project.Permission(user=template.manager)
            )
        template.update()
        return dump_data(schema, access=template)



import re
from dataclasses import dataclass
from math import cos, pi
from typing import Any

import bs4
import rasterio
import rasterio.errors

from ..exceptions import (
    CapabilitiesXMLParsingError,
    ExternalRequestError,
    UploadError,
)
from .requests import get
from .wfs import version2int

PREFERRED_FORMATS = ["geotiff", "tiff" "svg", "png"]


@dataclass
class WMSLayer:
    """This dataclass is meant for holding information parsed from
    a <Layer> XML tag in a WMS GetCapabilities query response.

    .. todo::
        Add other srs options from server, and use them to make the
        projection server-side if possible. Must be able to compare srs
        on multiple formats for that...
    """

    name: str
    title: str
    description: str
    keywords: list[str]
    styles: list[str]
    srs: list[str]
    bbox: list[tuple[str, float, float, float, float]]  # first str is CRS
    min_scale_denominator: float = 1
    max_scale_denominator: float = float("inf")
    parent: "WMSLayer" = None

    @property
    def default_style(self) -> str:
        """Default map style (first one from :attr:`styles` list)"""
        return self.styles[0]

    @property
    def queryable(self) -> bool:
        """Return ``True`` if layer has sufficient information to be
        queryable.

        Indeed because of layers tree structure, often root layer is a
        placeholder for default/common attributes but is not queryable.
        """
        return (
            self.name is not None
            and len(self.styles) > 0
            and len(self.srs) > 0
            and len(self.bbox) > 0
        )


@dataclass
class WMSCapabilities:
    """This dataclass holds information parsed from a WMS GetCapabilities query
    XML response.
    """

    url: str
    version: str
    layers: dict[str, WMSLayer]
    formats: list[str]

    @property
    def queryable_layers(self) -> dict[str, WMSLayer]:
        """Return dictionary of queryable layers"""
        return {k: v for k, v in self.layers.items() if v.queryable}


def convert_raw_crs(raw_crs: str) -> str:
    """Convert a crs/srs XML tag content into a usable srs string.

    :param raw_crs: crs/srs XML tag content
    :return: srs/csr usable for projections

    .. note::
        This function is here for future proofing this module.
        Indeed for now, raw crs are directly usable, but it may change with
        future WMS versions.
    """
    return raw_crs


def parse_layer(
    layer: bs4.Tag, version: str, parent: WMSLayer = None
) -> WMSLayer:
    """Parse a XML <Layer> tag.

    :param layer: XML <Layer> tag
    :param version: WMS version
    ;param parent: parent WMS layer if any
    :return: information describing the layer
    """
    version_int = version2int(version)
    title = "".join(layer.find("Title").contents)
    name_tag = layer.find("Name")
    if name_tag is not None:
        name = "".join(name_tag.contents)
    else:
        name = None
    description_tag = layer.find("Abstract")
    description = (
        "" if description_tag is None else "\n".join(description_tag.contents)
    )
    keywords = []
    keyword_iter = []
    if version_int >= (1, 1):
        keyword_tag = layer.find("KeywordList")
        if keyword_tag is not None:
            keyword_iter = keyword_tag.contents
    else:
        keyword_tag = layer.find("Keywords")
        if keyword_tag is not None:
            keyword_iter = ",".join(keyword_tag.contents).split(",")
    for key in keyword_iter:
        match key:
            case str():
                if len(re.findall(r"\S", key)) == 0:
                    continue
                keywords.append(key.strip())
            case bs4.Tag(name="Keyword"):
                keywords.append("".join(key.contents).strip())
    srs = [] if parent is None else parent.srs.copy()
    if version_int >= (1, 3):
        for crs_tag in layer.findAll("CRS"):
            srs.append(convert_raw_crs("".join(crs_tag.contents).strip()))
    else:
        for crs_tag in layer.findAll("SRS"):
            srs += re.findall(r"\S+", " ".join(layer.find("SRS").contents))
    styles = [] if parent is None else parent.styles.copy()
    for style_tag in layer.findAll("Style", recursive=False):
        styles.append("".join(style_tag.find("Name").contents))
    bbox = [] if parent is None else parent.bbox.copy()
    if version_int >= (1, 3):
        geobbox = layer.find("EX_GeographicBoundingBox")
        if geobbox is not None:
            bbox.insert(
                0,
                (
                    "EPSG:4326",
                    float(
                        "".join(geobbox.find("southBoundLatitude").contents)
                    ),
                    float(
                        "".join(geobbox.find("westBoundLongitude").contents)
                    ),
                    float(
                        "".join(geobbox.find("northBoundLatitude").contents)
                    ),
                    float(
                        "".join(geobbox.find("eastBoundLongitude").contents)
                    ),
                ),
            )
    elif version_int >= (1,):
        latlongbbox = layer.find("LatLonBoundingBox")
        if latlongbbox is not None:
            bbox.insert(
                0,
                (
                    "EPSG:4326",
                    float(latlongbbox.attrs.get("minx")),
                    float(latlongbbox.attrs.get("miny")),
                    float(latlongbbox.attrs.get("maxx")),
                    float(latlongbbox.attrs.get("maxy")),
                ),
            )
    for bbox_tag in layer.findAll("BoundingBox"):
        crs = bbox_tag.attrs.get("CRS" if version_int >= (1, 3) else "SRS")
        bbox.append(
            (
                crs,
                float(bbox_tag.attrs.get("minx")),
                float(bbox_tag.attrs.get("miny")),
                float(bbox_tag.attrs.get("maxx")),
                float(bbox_tag.attrs.get("maxy")),
            )
        )
    min_scale = None if parent is None else parent.min_scale_denominator
    max_scale = None if parent is None else parent.max_scale_denominator
    if version_int >= (1, 3):
        min_scale_tag = layer.find("MinScaleDenominator")
        if min_scale_tag is not None:
            min_scale = float("".join(min_scale_tag.contents))
        max_scale_tag = layer.find("MaxScaleDenominator")
        if max_scale_tag is not None:
            max_scale = float("".join(max_scale_tag.contents))

    return WMSLayer(
        name,
        title,
        description,
        keywords,
        styles,
        srs,
        bbox,
        min_scale,
        max_scale,
        parent=parent,
    )


def parse_layers_tree(
    layer_tag: bs4.Tag, version: str, parent: WMSLayer = None
) -> dict[str, WMSLayer]:
    """Parse a layer tag and all its subtags layers as well.

    :param layer_tag:
    :param version: WMS version
    :param parent:
        information about parent layer (some will be merged current layer to
        apply inherited attributes of parent), defaults to ``None``
    :return: information about all layers found in the tag tree
    """
    try:
        layer_info = parse_layer(layer_tag, version, parent)
    except (AttributeError, IndexError) as e:
        raise CapabilitiesXMLParsingError(str(e))
    res = {layer_info.name: layer_info}
    for subtag in layer_tag.find_all("Layer", recursive=False):
        res = {**res, **parse_layers_tree(subtag, version, layer_info)}
    return res


def parse_capabilities(
    capabilities: bs4.Tag,
) -> tuple[str, WMSCapabilities, list[str]]:
    """Parse a <WMS_Capabilities> XML tag for information.

    :param capabilities: XML tag
    :raises CapabilitiesXMLParsingError:
        if XML parsing of `capabilities` failed
    :return: WMS server capabilities
    """
    if capabilities is None:
        raise CapabilitiesXMLParsingError("cannot find tag 'Capabilities'")
    if not isinstance(capabilities, bs4.element.Tag):
        raise CapabilitiesXMLParsingError(
            f"Malformed GetCapabilities XML: 'Capabilities' should be a tag"
            f" instead of a {type(capabilities)}"
        )
    if "version" not in capabilities.attrs:
        raise CapabilitiesXMLParsingError(
            "Malformed GetCapabilities XML: cannot find 'version'"
        )
    version = capabilities.attrs.get("version")
    version_int = version2int(version)
    layers = {}
    for layer_tag in capabilities.find("Capability").find_all(
        "Layer", recursive=False
    ):
        layers = {**layers, **parse_layers_tree(layer_tag, version)}
    formats = []
    if version_int >= (1, 1):
        try:
            for format_tag in capabilities.find("GetMap").find_all("Format"):
                formats.append("".join(format_tag.contents))
        except AttributeError:
            raise CapabilitiesXMLParsingError("cannot find GetMap XML tag")
    else:
        try:
            for format_tag in (
                capabilities.find("Map")
                .find("Format")
                .find_all(lambda t: isinstance(t, bs4.Tag))
            ):
                formats.append("".join(format_tag.name.strip()))
        except AttributeError:
            raise CapabilitiesXMLParsingError("cannot find GetMap XML tag")

    return version, layers, formats


def get_capabilities(url: str, version: str = None) -> WMSCapabilities:
    """Perform WMS GetCapabilities request and parse informations.

    :param url: WMS server URL
    :param version:
        preferred WFS version number (actual is negotiated with server)
    :raises CapabilitiesXMLParsingError:
        if XML parsing of GetCapabilities response failed
    :raises ExternalRequestError:
        if GetCapabilities request to WMS server failed
    :return: WMS server capabilities
    """
    params = {
        "SERVICE": "WMS",
        "REQUEST": "GetCapabilities",
    }
    if version is not None:
        params["version"] = version
    response = get(url, params=params, request_type="GetCapabilities")
    soup = bs4.BeautifulSoup(response.content, features="xml")
    return WMSCapabilities(
        url,
        *parse_capabilities(soup.find(lambda tag: "Capabilities" in tag.name)),
    )


def choose_output_format(formats: list[str]) -> str:
    """Choose preferred available output format.

    :param formats: server available output formats
    :return: chosen one
    """
    _formats = [f.lower() for f in formats]

    for pf in PREFERRED_FORMATS:
        for f, format in zip(_formats, formats):
            if f.replace("image/", "") == pf:
                return format
    # Hail Mary: take first format and see if it will work
    return formats[0]


def bbox_to_size(
    wg84_bbox: tuple[float, float, float, float], resolution: float
) -> tuple[float, float]:
    """Compute ideal image size from EPSG:4236 bounding box.

    :param wg84_bbox: EPSG:4236 bounding box
    :param resolution: maximum size of a pixel (in meters)
    :return: image size
    """
    # PIXEL_SIZE = 0.00028  # resolution of one screen pixel in m
    EARTH_RADIUS = 6378137  # m
    _width_coeff = (wg84_bbox[3] - wg84_bbox[1]) * EARTH_RADIUS * pi / 180
    if wg84_bbox[2] * wg84_bbox[0] <= 0:
        width = _width_coeff
    else:
        width_up = _width_coeff * cos(wg84_bbox[2] * pi / 180)
        width_down = _width_coeff * cos(wg84_bbox[0] * pi / 180)
        width = max(width_down, width_up)
    height = (wg84_bbox[2] - wg84_bbox[0]) * EARTH_RADIUS * pi / 180
    return int(width / resolution), int(height / resolution)


def get_map(
    url: str,
    layer: str,
    version: str = None,
    bbox: tuple[float, float, float, float] = None,
    crs: str = None,
    style: str = None,
    format: str = None,
    size: tuple[float, float] = None,
    resolution: float = None,
    **kwargs,
) -> tuple[Any, Any]:
    """Perform a WMS GetMap request and retrieve result.

    :param url: WMS server URL
    :param layer:
    :param version: WMS version preferred (actual is negotiated with server)
    :param bbox:
        bounding box for image, defaults will take 'best' available bounding
        box (MUST also provide `crs` if set) from server
    :param crs:
        coordinate reference system in which to return image, defaults will
        take 'best' available coordinate reference system from server
    :param style:
        style to apply to image, defaults will take first one defined on server
    :param format:
        image output format, defaults will choose the most appropriate output
        format for this application
    :param size:
        size of image in pixels (width, height), defaults to None
    :param resolution:
        resolution of 1 pixel in meters, defaults to 1000.0
        (unused if `size` argument is set)
    :raises KeyError:
        if we cannot find a server defined bounding box for user provided `crs`
    :raises ValueError:
        if `size` was not provided and cannot be inferred
    :raises UploadError:
        if the GetMap response could not be loaded into geodata
    :raises ExternalRequestError: if GetMap request to WMS server failed
    :raises CapabilitiesXMLParsingError:
        if XML parsing of GetCapabilities failed
    :return: tuple with first result image then source object
    """
    resolution = resolution or 1000.0
    capabilities = get_capabilities(url, version)
    layer_data = capabilities.layers[layer]
    version = capabilities.version
    version_int = version2int(version)
    format = format or choose_output_format(capabilities.formats)
    style = style or layer_data.default_style
    if crs is None:
        for bbox_tuple in layer_data.bbox:
            if bbox_tuple[0] in layer_data.srs:
                crs = bbox_tuple[0]
                bbox = bbox_tuple[1:]
                break
    if bbox is None:
        for bbox_tuple in layer_data.bbox:
            if bbox_tuple[0] == crs:
                bbox = bbox_tuple[1:]
                break
    if bbox is None:
        raise KeyError(
            f"No BoundingBox provided for crs {crs} and none are defined "
            "on server"
        )
    params = {
        "SERVICE": "WMS",
        "LAYERS": [layer],
        "BBOX": ",".join(str(p) for p in bbox),
        "FORMAT": format,
        "STYLES": [style],
    }

    if version_int >= (1, 3):
        params["CRS"] = crs
    else:
        params["SRS"] = crs
    if version_int >= (1, 1):
        params["VERSION"] = version
        params["REQUEST"] = "GetMap"
    else:
        params["WMTVER"] = version
        params["REQUEST"] = "Map"
    for k, v in kwargs.items():
        params[k.capitalize()] = v
    if size is None:
        for bbox_tuple in layer_data.bbox:
            if bbox_tuple[0] == "EPSG:4326":
                wgs84_bbox = bbox_tuple[1:]
                size = bbox_to_size(wgs84_bbox, resolution)
                break
    if size is None:
        raise ValueError(
            "Argument 'size' could not be inferred and was not provided"
        )
    params["WIDTH"] = size[0]
    params["HEIGHT"] = size[1]
    response = get(url, params=params, request_type="GetMap")
    try:
        with rasterio.MemoryFile(response.content) as memfile:
            with memfile.open(crs=crs) as src:
                img = src.read(1, masked=True)
                return img, src
    except rasterio.errors.RasterioIOError:
        try:
            if "xml" not in response.headers.get("content-type", "").lower():
                raise UploadError(
                    f"Cannot load geodata from URL: {response.url}"
                )
            soup = bs4.BeautifulSoup(response.content, features="xml")
            etag = soup.find("ServiceExceptionReport").find("ServiceException")
            srv_msg = f"{etag.attrs['code']}: " + "".join(
                etag.contents
            ).replace("\n", "")
        except (IndexError, AttributeError):
            raise UploadError(f"Cannot load geodata from URL: {response.url}")
        raise ExternalRequestError(
            response.url,
            f"WMS server cannot return geodata. {srv_msg}",
            request_type="GetMap",
        )



from copy import copy, deepcopy

from geoalchemy2.shape import from_shape
from geoalchemy2.types import Geometry
from shapely.geometry import box
from sqlalchemy.ext.hybrid import hybrid_property

from . import db
from .mixins import TimestampMixin
from .permission import HasPermissions


class ProjectBase(TimestampMixin, HasPermissions, db.Model):
    __tablename__ = "project"

    is_template = db.Column(db.Boolean, nullable=False)
    name = db.Column(db.String, nullable=False)
    manager_id = db.Column(db.Integer, db.ForeignKey("user.id"))
    description = db.Column(db.String)
    extent = db.Column(Geometry, nullable=False)

    _manager = db.relationship("User")
    data_list = db.relationship(
        "ProjectData",
        cascade="all,delete-orphan",
        backref="project",
        primaryjoin="ProjectData.project_id == Project.id",
    )

    __mapper_args__ = {"polymorphic_on": is_template}

    def __init__(
        self, name, manager=None, description="", extent=None, is_public=False
    ):
        self.name = name
        self.description = description
        self.manager = manager
        if extent is None:
            shp_box = box(-180.0, -90.0, 180.0, 90.0)
            extent = from_shape(shp_box, srid=4326)
        self.extent = extent
        self.is_public = is_public

    @hybrid_property
    def manager(self):
        return self._manager

    @manager.setter
    def manager(self, user):
        self._manager = user
        if user is not None:
            self.permissions.append(ProjectBase.Permission(user=user))


class Template(ProjectBase):
    """
    Represents the template of a project.

    A template does not have permissions.
    Only the manager of the template should have access to it.
    """

    __mapper_args__ = {"polymorphic_identity": True}
    __tablename__ = None

    def __repr__(self):
        return "<Template(name={})>".format(self.name)

    @classmethod
    def from_project(cls, project, manager=None):
        """Create a new project template from an existing project."""
        manager = manager or project.manager
        template = cls(
            name=project.name,
            manager=manager,
            description=project.description,
            extent=copy(project.extent),
            is_public=project.is_public,
        )
        # memo used by deepcopy
        # it would here consider that the references to the project
        # are in fact references to the template
        memo = {id(project): template}
        template.data_list = [deepcopy(d, memo) for d in project.data_list]
        return template


class Project(ProjectBase):
    """Represents a project and its data"""

    __mapper_args__ = {"polymorphic_identity": False}
    __tablename__ = None

    def __repr__(self):
        return "<Project(name={}, manager_id={}>".format(
            self.name, self.manager_id
        )

    @classmethod
    def from_template(cls, template, manager=None):
        """
        Create a project from a template.

        The manager of the project will be the owner of the template.

        Args:
            template: the template to copy
        """
        manager = manager or template.manager
        project = cls(
            name=template.name,
            manager=manager,
            description=template.description,
            extent=copy(template.extent),
        )
        memo = {id(template): project}
        project.data_list = [deepcopy(d, memo) for d in template.data_list]
        return project



from flask import Flask

from . import celery as celery_ext
from . import migrate as migrate_ext
from . import swagger as swagger_ext


def init_app(app: Flask):
    swagger_ext.init_app(app)
    celery_ext.init_app(app)
    migrate_ext.init_app(app)



from celery.result import AsyncResult
from flask import jsonify
from flask_restful import Resource, reqparse

from ..exceptions import RequestError
from ..models import ProjectTaskModel
from ..schemas import (
    MessageSchema,
    ProjectTaskListGetResponseSchema,
    ProjectTaskSchema,
    TaskGetResponseSchema,
)
from ..services import project_service
from ..services.auth_service import token_required
from .utils import check_permission, dump_data, with_query_arg, with_response


class ProjectTaskListAPI(Resource):
    def get_tasks_list_schemas(self, project_id, limit=None, offset=None):
        """Returns the list of tasks concerning the data of a project"""
        task_models = project_service.get_project_tasks(
            project_id, limit=limit, offset=offset
        )
        schema = ProjectTaskSchema(many=True)
        data = {
            "tasks": schema.dump(task_models),
            "total": project_service.get_total_project_tasks(project_id),
        }
        return jsonify(**data)

    @token_required
    @with_response(
        200,
        schema=ProjectTaskListGetResponseSchema,
        description="List of project tasks",
    )
    @with_query_arg(
        "limit",
        int,
        required=False,
        description="Maximum number of tasks to return. Default: all",
    )
    @with_query_arg(
        "offset",
        int,
        required=False,
        description="Number of tasks to skip. Default: None",
    )
    def get(self, project_id):
        """
        Return the list of tasks for the project

        :param int project_id: The id of the project.
        :query int limit: Number of tasks to return.
        :query int offset:
            Number of tasks to skip (e.g., a value of ``25`` means the first 25
            tasks will not be returned).
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The project doesn't exist.

        .. :quickref: Project; Get the list of tasks in a project
        """
        project_service.get_if_authorized(project_id)
        parser = reqparse.RequestParser()
        parser.add_argument("limit", type=int, location="args")
        parser.add_argument("offset", type=int, location="args")
        args = parser.parse_args()
        limit = args.get("limit", None)
        offset = args.get("offset", None)
        return self.get_tasks_list_schemas(
            project_id, limit=limit, offset=offset
        )


class TaskAPI(Resource):
    @token_required
    @with_response(
        200, schema=TaskGetResponseSchema, description="Details of the task"
    )
    def get(self, task_id):
        """
        Return the details of the task.

        :param int task_id: The id of the task.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access the project for this task.
        :status 404: The task doesn't exist.

        .. :quickref: Task; Get the details of the task
        """
        task = ProjectTaskModel.get_by_task_id(task_id)
        if task is None:
            raise RequestError("Task not found", 404)
        check_permission(task.project)
        return dump_data(ProjectTaskSchema(), task=task)

    @token_required
    @with_response(
        200, schema=MessageSchema, description="Task revoked successfully"
    )
    def delete(self, task_id):
        """
        Revoke the task.

        :param int task_id: The id of the task.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access the project for this task.
        :status 404: The task doesn't exist.

        .. :quickref: Task; Revoke the task
        """
        task = ProjectTaskModel.get_by_task_id(task_id)
        if task is None:
            raise RequestError(f"Task {task_id} does not exist", 404)
        check_permission(task.project)
        result = AsyncResult(task_id)
        result.revoke(terminate=True)
        return jsonify(message="task {} deleted".format(task_id))



from enum import Enum

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import backref

from . import db
from .mixins import ModelMixin


def find_tablename(cls):
    """Find `__tablename__` attribute in class tree.

    :raises AttributeError: if attribute `__tablename__` is never set in tree
    :return: closest set `__tablename__` attribute found
    """
    if cls.__tablename__ is None:
        for scls in cls.__bases__:
            if issubclass(scls, db.Model):
                return find_tablename(scls)
        raise AttributeError(f"cannot find {cls} __tablename__")
    return cls.__tablename__


class HasPermissions:
    """Mixin to create a permission table for the object

    Inspired from SQLAlchemy's example:
    http://docs.sqlalchemy.org/en/rel_1_1/_modules/examples/generic_associations/table_per_related.html
    """  # E501

    is_public = db.Column(db.Boolean, default=False, nullable=False)

    @declared_attr
    def permissions(cls):
        """Create new a permissions table linking a user with an object

        The table created will be called '<name>_permission', with <name> being
        the name of the table where this mixin is used
        """
        cls.Permission = type(
            "{}Permission".format(cls.__name__),
            (ModelMixin, db.Model),
            dict(
                __tablename__="{}_permission".format(find_tablename(cls)),
                object_id=db.Column(
                    db.Integer,
                    db.ForeignKey("{}.id".format(find_tablename(cls))),
                ),
                object=db.relationship(cls),
                user_id=db.Column(
                    db.Integer, db.ForeignKey("user.id"), nullable=False
                ),
                user=db.relationship("User"),
            ),
        )
        return db.relationship(cls.Permission, cascade="all,delete-orphan")

    def is_user_authorized(self, user):
        """Check if the user has a permission on this object

        Returns a boolean indicating if the user has a permission on the object
        """
        return self.is_public or any(
            (user.id == perm.user_id for perm in self.permissions)
        )


class PermissionAbility(Enum):
    create_project = 1
    create_geo_data = 2


class UserPermission(ModelMixin, db.Model):
    __tablename__ = "user_permission"

    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    ability = db.Column(db.Enum(PermissionAbility), nullable=False)

    user = db.relationship(
        "User",
        backref=backref(
            "permissions", lazy="select", cascade="all,delete-orphan"
        ),
    )



import io
import json
import os
import subprocess as sp
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Generator
from uuid import uuid4
from zipfile import ZipFile, is_zipfile

import fiona
import fiona.errors
import geopandas as gpd
import numpy as np
import pyproj
import pyproj.crs
import rasterio
import yaml
from fiona.crs import from_epsg
from flask import current_app as app
from geoalchemy2.shape import from_shape, to_shape
from rasterio.features import shapes
from scipy import ndimage
from shapely.geometry import LinearRing, LineString, Point, Polygon, shape
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from shapely.ops import transform
from shapely.prepared import prep
from werkzeug.utils import secure_filename

from ..exceptions import RequestError, UploadError
from ..models import db
from ..models.geo_data import (
    BaseGeoData,
    DataAttribute,
    DataAttributeNominal,
    DataAttributeQuantitative,
    Feature,
    GeoData,
    GeoFeature,
    GlobalData,
    WFSGeoData,
    WMSGeoData,
)
from . import wfs, wms

DEFAULT_SRID = 4326
DEFAULT_PROJECTION = f"epsg:{DEFAULT_SRID}"
DEFAULT_DRIVER = "ESRI Shapefile"


def get_temporary_directory():
    """
    Return a TemporaryDirectory object.

    If the 'TEMP_DIR' defined in the application configuration exists,
    the TemporaryDirectory object will use this location. Otherwise,
    use the system temporary directory.

    Returns:
        A file object directory which can be used in a context manager and
        will be deleted on completion of the context.
    """
    temp_dir_conf = app.config.get("TEMP_DIR", None)
    if temp_dir_conf is None or os.path.isdir(temp_dir_conf) is False:
        temp_dir = None
    else:
        temp_dir = temp_dir_conf
    return TemporaryDirectory(dir=temp_dir)


def convert_to_polygon(geom: BaseGeometry) -> Polygon:
    """Try converting geometry to a polygon.

    :param geom:
    :raises TypeError:
        if `geom` cannot be converted to a :class:`shapely.Polygon`
    :return: result as a polygon
    """
    match geom:
        case Polygon():
            return geom
        case Point():
            return Polygon(list(geom.coords) * 4)
        case LinearRing():
            return Polygon(geom.coords)
        case LineString():
            coords = list(geom.coords)
            return Polygon(coords + coords[::-1])
        case _:
            raise TypeError(
                f"{type(geom).__name__} cannot be converted to Polygon"
            )


def split_multi_geometry(
    geom: BaseGeometry,
) -> Generator[BaseGeometry, None, None]:
    """Return list of connex geometries forming the geometry.

    :param geom: a geometry
    :return:
    """
    match geom:
        case BaseMultipartGeometry():
            for g in geom.geoms:
                yield from split_multi_geometry(g)
        case _:
            yield geom


def post_load_normalization(data: BaseGeoData):
    """Split multi-geometry features of persisted base geo data.

    The resulting features are then persisted back to the database.

    :param data: persisted base geo data object

    .. warning::
        This is intended to be used immedialetly after data features creation.
        Only values and input features are preserved, output features are not!
    """
    memo = {}
    memo[id(data)] = data
    for attr in data.attributes:
        memo[id(attr)] = attr
    features_list = []
    for feature in data.features:
        geom = to_shape(feature.geom)
        parts = list(split_multi_geometry(geom))
        if len(parts) > 1:
            feature.geom = from_shape(parts.pop(0), srid=DEFAULT_SRID)
            features_list.append(feature)
            for g in parts:
                new_feature = deepcopy(feature, memo)
                new_feature.geom = from_shape(g, srid=DEFAULT_SRID)
                new_feature.input_features = feature.input_features
                features_list.append(new_feature)
    db.session.add_all(features_list)
    db.session.commit()


def normalize_geometry(df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Normalize geopandas dataframe geometry type.

    If geometry column has more than 1 type, then it is converted to polygons.
    Otherwise it is not changed.

    :param df: geopandas dataframe
    :return: normalized geopandas dataframe

    .. warning::
        Dataframe is modified in place! It is also returned for chainability
        of methods or functions.
    """
    geom_types = gpd.io.file.infer_schema(df)["geometry"]
    if isinstance(geom_types, str):
        geom_types = [geom_types]
    if len(geom_types) == 1:
        return df

    df.geometry = df.geometry.apply(convert_to_polygon)
    return df


def create_value_classes(
    start: float, stop: float, step: float
) -> list[list[float]]:
    """Create value classes from range.

    This is to be used to discretize raster geodata values so to create larger
    vector features sharing a same value class.
    Each class is a list with `[start_value, stop_value, class_value]`
    (i.e values between `start_value` and `stop_value` are replaced by
    `class_value`).

    :param start:
    :param stop:
    :param step:
    :return: value classes
    """
    range_ = np.arange(start, stop, step).tolist()
    if range_[-1] < stop:
        range_.append(stop)
    return [
        [_start, _stop, _start]
        for _start, _stop in zip(range_[:-1], range_[1:])
    ]


def save_geo_data(geo_data: BaseGeoData, directory, file_ext=None):
    """
    Save a geo data to the directory.

    It also adds following metadata as geo data fields:
    * explainability: feature explainability if any

    Returns:
        The name of the new directory.
    """
    properties = []
    for feature in geo_data.features:
        prop = {
            p.attribute.name: p.value
            for p in feature.properties
            if p.attribute.name != "fid"
        }
        prop["explainability"] = json.dumps(feature.explain())
        properties.append(prop)
    df = gpd.GeoDataFrame(
        properties,
        geometry=gpd.GeoSeries([to_shape(f.geom) for f in geo_data.features]),
        crs=from_epsg(DEFAULT_SRID),
    )

    # Normalize geometry to polygons if more than 1 type
    df = normalize_geometry(df)

    # Convert the booleans to int (booleans are not valid shapefile values)
    # see https://github.com/geopandas/geopandas/issues/437
    coltypes = gpd.io.file.infer_schema(df)["properties"]
    for colname, coltype in coltypes.items():
        if coltype == "bool":
            app.logger.info(
                "bool converted as int in {} while saving the data {}"
                "".format(colname, geo_data.name)
            )
            df[colname] = df[colname].astype("int")

    filename = secure_filename(geo_data.name)
    if file_ext is not None and file_ext.lower() != "zip":
        filename += "." + file_ext
    path = os.path.join(directory, filename)
    try:
        df.to_file(path)
    except fiona.errors.GeometryTypeValidationError:
        raise RequestError("Error in the geometries types", 500)
    except fiona.errors.DriverError:
        raise RequestError(f"File format unsupported '{file_ext}'", 415)
    return filename


def save_global_data(global_data: GlobalData, directory, file_ext=None):
    file_ext = file_ext or "json"
    properties = {p.attribute.name: p.value for p in global_data.properties}
    filename = secure_filename(global_data.name) + "." + file_ext
    path = os.path.join(directory, filename)
    with open(path, "w") as f:
        if file_ext == "json":
            json.dump(properties, f)
        elif file_ext == "yaml" or file_ext == "yml":
            yaml.dump(properties, f)
        else:
            raise RequestError(f"File format unsupported '{file_ext}'", 415)
    return filename


def get_geo_data_download(data, working_dir, file_ext=None):
    """Write the geo data to a file-like object."""
    match data:
        case BaseGeoData():
            filename = save_geo_data(data, working_dir, file_ext)
        case GlobalData():
            filename = save_global_data(data, working_dir, file_ext)
    path = os.path.join(working_dir, filename)
    print(filename, path)
    print(os.path.isfile(path), os.path.isdir(path))
    if os.path.isfile(path):
        return filename, path
    elif os.path.isdir(path):
        buff = io.BytesIO()
        with ZipFile(buff, "w") as zf:
            files = os.listdir(path)
            for f in files:
                zf.write(os.path.join(path, f), os.path.join(filename, f))
        buff.flush()
        buff.seek(0)
        return filename + ".zip", buff


def ogr2ogr_reproject(file_in, dir_out):
    """
    Reproject the file in WGS84 using ogr2ogr and return the name of the new
    file.

    Args:
        file_in (str): Path of the input shapefile.
        dir_out (str): Directory to save the reprojected file.

    Returns:
        The path of the reprojected file, or None if there was an error.
    """
    ogr2ogr_bin = app.config.get("OGR2OGR_BIN", None)
    if ogr2ogr_bin is None:
        return None
    if os.path.isfile(ogr2ogr_bin) is False:
        app.logger.warning("ogr2ogr not found at {}".format(ogr2ogr_bin))
        return None
    out_filename = str(uuid4())
    file_out = os.path.join(dir_out, out_filename)
    args = [
        ogr2ogr_bin,
        "-f",
        DEFAULT_DRIVER,
        file_out,
        file_in,
        "-t_srs",
        DEFAULT_PROJECTION.upper(),
    ]
    try:
        sp.check_call(args)
        return out_filename
    except FileNotFoundError:
        return None
    except sp.CalledProcessError:
        return None


def validate_shp_filenames(filenames):
    """
    Validate the file names.

    A shapefile data must have at least 3 files with the extensions .shp, .shx
    and .dbf.

    Raises:
        UploadError:
            if a required file (shp, dbf or shx) is missing, or if
            more than one file is found
    """
    if len(filenames) < 3:
        raise UploadError(
            "The shapefile should contain at least 3 files: "
            ".shp, .shx and .dbf"
        )
    f_exts = []
    for f in filenames:
        _, f_ext = os.path.splitext(f)
        f_exts.append(f_ext.lower())
    required = (".shp", ".shx", ".dbf")
    missing = list(filter(lambda x: x not in f_exts, required))
    if len(missing) > 0:
        raise UploadError(
            "Missing files extensions: {}".format(", ".join(missing))
        )


def create_extent_filter(extent_filter: BaseGeometry = None):
    """Create a filter function which returns True if a geometry is
    filtered out.

    :param extent_filter:
        geometry used to filter in geometries, defaults to None
    :return: filter boolean function
    """
    if extent_filter is None:
        return lambda g: False
    extent_prep = prep(extent_filter)
    return lambda g, e=extent_prep: not e.intersects(g)


def create_global_data(
    name, properties: dict[str, Any], **kwargs
) -> GlobalData:
    """Create a global data object.

    :param name:
    :param properties: set of properties organized by attribute
    :return: created global data object
    """
    data = GlobalData(name=name, feature=None, **kwargs)
    data.update_created()
    db.session.add(data)
    feature = Feature(data=data)
    data.feature = feature
    db.session.add(feature)
    values = []
    for aname, avalue in properties.items():
        attr = data._choose_attribute_type(avalue)(name=aname, data=data)
        values.append(
            attr.get_value_class()(
                value=avalue, attribute=attr, feature=feature
            )
        )
        data.attributes.append(attr)
    db.session.add_all(values)
    db.session.commit()
    return data


def load_vector_file(path, extent_filter: BaseGeometry = None):
    """
    Create and return a GeoData object created from a vector file.

    Tries to reproject the geometries to EPSG:4326.
    """
    with fiona.open(path, "r") as source:
        project = None
        if source.crs != {}:
            proj_out = pyproj.crs.CRS(DEFAULT_PROJECTION)
            proj_in = pyproj.crs.CRS(source.crs)
            if not proj_in.equals(proj_out):
                project = pyproj.Transformer.from_crs(
                    proj_in, proj_out
                ).transform
        geo_data = GeoData(
            name=source.name,
            original_name=source.name,
            source_driver=source.driver,
        )
        geo_data.update_created()
        # Generate an id for the geo data
        db.session.add(geo_data)
        db.session.flush()
        features = []
        values = []
        for aname, atype in source.schema["properties"].items():
            proptype = fiona.prop_type(atype)
            if proptype in (int, float):
                attr = DataAttributeQuantitative(
                    name=aname, data_id=geo_data.id
                )
            else:
                attr = DataAttributeNominal(name=aname, data_id=geo_data.id)
            geo_data.attributes.append(attr)
        attributes_dict = {a.name: a for a in geo_data.attributes}
        extent_filter_fn = create_extent_filter(extent_filter)
        for s_feature in source:
            geom = shape(s_feature["geometry"])
            if project is not None:
                geom = transform(project, geom)
            if extent_filter_fn(geom):
                # Ignore geometry outside extent (if set)
                continue
            geom_geometry = from_shape(geom, srid=DEFAULT_SRID)
            feature = GeoFeature(geom=geom_geometry, data_id=geo_data.id)
            geo_data.features.append(feature)
            props = s_feature["properties"]
            for prop, val in props.items():
                cls = attributes_dict[prop].get_value_class()
                values.append(
                    cls(
                        value=val,
                        attribute=attributes_dict[prop],
                        feature=feature,
                    )
                )
        db.session.add_all(values)
        db.session.add_all(features)
        db.session.commit()
        return geo_data


def load_wfs(wfs_stream: WFSGeoData) -> WFSGeoData:
    """Create and return a StreamGeoData object created from a WFS URL.

    :param wfs_stream:
        WFS stream to load (can be partial as long as it contains at least
        the server URL and feature type)
    :raise KeyError: if `wfs_stream.feature_type` is not found on WFS server
    :return:

    .. note:: WFS object is not persisted!
    """
    capabilities = wfs.get_capabilities(wfs_stream.url, wfs_stream.version)
    if wfs_stream.feature_type not in capabilities.feature_types:
        raise KeyError(
            f"No feature type found on server {wfs_stream.url} with given "
            f"name '{wfs_stream.feature_type}'"
        )
    feature_info = capabilities.feature_types[wfs_stream.feature_type]
    # Do we really need original_name field?
    wfs_stream.original_name = wfs_stream.feature_type
    wfs_stream.name = secure_filename(feature_info.name)
    wfs_stream.title = feature_info.title
    wfs_stream.description = feature_info.description
    wfs_stream.keywords = feature_info.keywords

    # Load features so stream can be viewed directly
    features = load_from_wfs(wfs_stream)
    # List the data attributes
    attrs = []
    for f in features:
        for p in f.properties:
            if p.attribute not in attrs:
                attrs.append(p.attribute)
    wfs_stream.features = features
    wfs_stream.attributes = attrs

    return wfs_stream


def load_from_wfs(
    wfs_stream: WFSGeoData,
    old_attributes: list[DataAttribute] = None,
) -> list[GeoFeature]:
    """Create and return a GeoData object created from a WFS stream.

    Tries to reproject the geometries to EPSG:4326.

    :param wfs_stream:
    :return: geo data version of the `wfs_stream` stream
    """
    df = wfs.get_feature(
        wfs_stream.url, wfs_stream.feature_type, wfs_stream.version
    )
    if df.crs:
        df = df.to_crs(DEFAULT_PROJECTION)
    old_attributes = old_attributes or []
    old_attrs_dict = {attr.name: attr for attr in old_attributes}
    attributes_dict = {}
    features = []
    for aname, atype in df.dtypes.items():
        if aname == df.geometry.name:
            continue
        if atype in (float, int):
            attr = DataAttributeQuantitative(name=aname)
        else:
            attr = DataAttributeNominal(name=aname)
        old_attr = old_attrs_dict.get(aname, attr)
        attributes_dict[aname] = old_attr if attr.same_as(old_attr) else attr

    attr_values = defaultdict(list)
    extent_filter = (
        None
        if wfs_stream.extent_filter is None
        else to_shape(wfs_stream.extent_filter)
    )
    extent_filter_fn = create_extent_filter(extent_filter)
    for s_feature, geom in zip(df.iterfeatures(), df.geometry.values):
        if extent_filter_fn(geom):
            # Ignore geometry outside extent (if set)
            continue
        geom_geometry = from_shape(geom, srid=DEFAULT_SRID)
        feature = GeoFeature(geom=geom_geometry)
        features.append(feature)
        for prop, val in s_feature["properties"].items():
            cls = attributes_dict[prop].get_value_class()
            attr_values[attributes_dict[prop]].append(
                cls(
                    value=val,
                    attribute=attributes_dict[prop],
                    feature=feature,
                )
            )

    return features


def load_vector_from_files(files, extent_filter: BaseGeometry = None):
    """
    Create a new GeoData from a list of files.

    Saves the files in a temporary directory and load them as a new GeoData
    object.
    """
    with get_temporary_directory() as temp_dir:
        temp_files = []
        for f in files:
            # This file datastructure can be read only once
            # Without it the is_zipfile test seem to empty it for later cases
            # so other file type are not read properly
            path = os.path.join(temp_dir, secure_filename(f.filename))
            f.save(path)
            temp_files.append(path)
        if len(temp_files) == 1 and is_zipfile(temp_files[0]):
            # Pre-treatment on zipped shapefile
            file = temp_files.pop(0)
            with ZipFile(file) as zip_file:
                zip_infos = []
                for zi in zip_file.infolist():
                    basename = os.path.basename(zi.filename)
                    if basename != "":
                        zi.filename = secure_filename(basename)
                        zip_infos.append(zi)
                for zi in zip_infos:
                    temp_files.append(zip_file.extract(zi, path=temp_dir))
        if len(temp_files) > 1:
            # Shapefile
            validate_shp_filenames(temp_files)
            shape_file = None
            for f_name in temp_files:
                f_root, f_ext = os.path.splitext(f_name)
                if f_ext == ".shp":
                    # Stop at the first .shp file
                    shape_file = f_name
                    break
            if shape_file is None:
                raise UploadError("No file .shp found.")
            file_in = shape_file
        else:
            # Any other format
            file_in = temp_files[0]

        out_filename = ogr2ogr_reproject(file_in, temp_dir)
        fname = out_filename if out_filename is not None else file_in
        try:
            return load_vector_file(
                os.path.join(temp_dir, fname), extent_filter
            )
        except fiona.errors.DriverError as exc:
            message = str(exc).replace(
                os.path.join(temp_dir, fname), Path(file_in).suffix
            )
            raise UploadError(message)


def reclassify(array, classes=None, start=None, step=None, stop=None):
    """Reclassify the raster values."""
    if classes is None:
        _flattened = array[~np.isnan(array)].data.flatten()
        start = min(_flattened) if start is None else start
        stop = max(_flattened) if stop is None else stop
        step = step or (stop - start) / 100
        classes = create_value_classes(start, stop, step)

    result = np.ma.masked_invalid(array)
    min_value, max_value = float("inf"), -float("inf")
    min_label, max_label = None, None
    for c in classes:
        v_min, v_max = sorted((c[0], c[1]))
        v_val = c[2]
        if v_min < min_value:
            min_value = v_min
            min_label = v_val
        if v_max > max_value:
            max_value = v_max
            max_label = v_val
        result.data[
            np.where((array >= v_min) & (array < v_max) & ~result.mask)
        ] = v_val
    result.data[np.where((array < min_value) & ~result.mask)] = min_label
    result.data[np.where((array >= max_value) & ~result.mask)] = max_label
    return result


def load_from_geotiff(
    path,
    classes,
    start: float = None,
    step: float = None,
    stop: float = None,
    smooth=False,
    extent_filter: BaseGeometry = None,
):
    """Create a new geo-data from a GeoTiff file.

    Args:
        path (str): The path of the GeoTIFF file.
        classes (list):
            The classes of the vector features, a list of (min, max, value).
        start (float):
            The starting value for classes creation
            (if `classes` not provided).
        step (float):
            The step value for classes creation (if `classes` not provided).
        stop (float): The stop value for classes (if `classes` not provided).
        smooth (bool):
            Whether to smooth the raster using a percentile_filter.
            Default: False.
        extent_filter (shapely.BaseGeometry):
            Geometry filtering in features to keep.
            Default: None (no filter).
    """
    with rasterio.open(path) as src:
        image = src.read(1, masked=True)

    name = src.name
    name = os.path.basename(name)
    name = os.path.splitext(name)[0]
    features = _load_features_from_raster(
        image,
        src,
        classes=classes,
        start=start,
        step=step,
        stop=stop,
        smooth=smooth,
        extent_filter=extent_filter,
    )
    # List the data attributes
    attrs = []
    for f in features:
        for p in f.properties:
            if p.attribute not in attrs:
                attrs.append(p.attribute)
    geo_data = GeoData(
        name=name,
        original_name=name,
        source_driver=src.driver,
        features=features,
        attributes=attrs,
    )
    geo_data.create()
    return geo_data


def load_wms(wms_stream: WMSGeoData) -> WMSGeoData:
    """Create and return a StreamGeoData object created from a WFS URL.

    :param wms_stream:
        WMS stream to load (can be partial as long as it contains at least
        the server URL and layer)
    :raise KeyError: if `wms_stream.layer` is not found on WMS server
    :return:

    .. note:: WMS object is not persisted!
    """
    capabilities = wms.get_capabilities(wms_stream.url, wms_stream.version)
    if wms_stream.layer not in capabilities.queryable_layers:
        raise KeyError(
            f"No queryable layer found on server {wms_stream.url} "
            f"with given name '{wms_stream.layer}'"
        )
    layer_info = capabilities.layers[wms_stream.layer]
    # Do we really need original_name field?
    wms_stream.original_name = wms_stream.layer
    wms_stream.name = secure_filename(layer_info.name)
    wms_stream.title = layer_info.title
    wms_stream.description = layer_info.description
    wms_stream.keywords = layer_info.keywords

    # Load features so stream can be viewed directly
    features = load_from_wms(
        wms_stream,
        classes=wms_stream.classes,
        start=wms_stream.start,
        step=wms_stream.step,
        stop=wms_stream.stop,
        resolution=wms_stream.resolution,
    )
    # List the data attributes
    attrs = []
    for f in features:
        for p in f.properties:
            if p.attribute not in attrs:
                attrs.append(p.attribute)
    wms_stream.features = features
    wms_stream.attributes = attrs

    return wms_stream


def load_from_wms(
    wms_stream: WMSGeoData,
    classes: list = None,
    start: float = None,
    step: float = None,
    stop: float = None,
    smooth: bool = False,
    resolution: float = None,
    old_attribute: DataAttribute = None,
) -> list[GeoFeature]:
    """Create a new geo-data from a WMS url.

    :param wms_stream: The WMS stream.
    :param classes:
        The classes of the vector features, a list of (min, max, value).
    :param start:
        The starting value for classes creation (if `classes` not provided).
    :param step:
        The step value for classes creation (if `classes` not provided).
    :param stop: The stop value for classes (if `classes` not provided).
    :param smooth:
        Whether to smooth the raster using a percentile_filter.
        Default: False.
    """
    img, src = wms.get_map(
        wms_stream.url,
        wms_stream.layer,
        wms_stream.version,
        resolution=resolution,
    )
    extent_filter = (
        None
        if wms_stream.extent_filter is None
        else to_shape(wms_stream.extent_filter)
    )
    geodata = _load_features_from_raster(
        img,
        src,
        classes=classes,
        start=start,
        step=step,
        stop=stop,
        smooth=smooth,
        old_attribute=old_attribute,
        extent_filter=extent_filter,
    )
    return geodata


def _load_features_from_raster(
    image,
    src,
    classes: list[list[float]] = None,
    start: float = None,
    step: float = None,
    stop: float = None,
    smooth: bool = False,
    old_attribute: DataAttribute = None,
    extent_filter: BaseGeometry = None,
) -> list[GeoFeature]:
    """Load GeoFeatures from raster image.

    Args:
        image: the loaded raster image (previously read from `src`)
        src: dataset reader
        classes (list):
            The classes of the vector features, a list of (min, max, value).
        start (float):
            The starting value for classes creation
            (if `classes` not provided).
        step (float):
            The step value for classes creation (if `classes` not provided).
        stop (float):
            The stop value for classes (if `classes` not provided).
        smooth (bool):
            Whether to smooth the raster using a percentile_filter.
            Default: False.
        extent_filter (shapely.BaseGeometry):
            Geometry filtering in features to keep.
            Default: None (no filter).
    """
    image = image.astype(np.float32)
    image[np.where(image.mask)] = np.nan
    image = reclassify(image, classes, start=start, step=step, stop=stop)

    if smooth:
        filtered = ndimage.percentile_filter(image, percentile=25, size=7)
        np.copyto(image.data, filtered, where=(~image.mask))

    results = [
        {"value": v, "geometry": s}
        for i, (s, v) in enumerate(
            shapes(
                image,
                mask=(~image.mask & np.isfinite(image.data)),
                transform=src.transform,
            )
        )
    ]

    project = None
    if src.crs != {}:
        proj_out = pyproj.crs.CRS(DEFAULT_PROJECTION)
        proj_in = pyproj.crs.CRS(src.crs)
        if not proj_in.equals(proj_out):
            project = pyproj.Transformer.from_crs(proj_in, proj_out).transform

    if all(
        (
            isinstance(r["value"], int) or isinstance(r["value"], float)
            for r in results
        )
    ):
        attr = DataAttributeQuantitative(name="value")
    else:
        attr = DataAttributeNominal(name="value")
    if old_attribute is not None:
        attr = old_attribute if attr.same_as(old_attribute) else attr

    features = []
    values = []
    val_cls = attr.get_value_class()
    extent_filter_fn = create_extent_filter(extent_filter)
    for result in results:
        geom = shape(result["geometry"])
        if project is not None:
            geom = transform(project, geom)
        if extent_filter_fn(geom):
            # Ignore geometry outside extent (if set)
            continue
        geom_geometry = from_shape(geom, srid=DEFAULT_SRID)
        feature = GeoFeature(geom=geom_geometry)
        values.append(
            val_cls(value=result["value"], attribute=attr, feature=feature)
        )
        features.append(feature)
    return features


def load_raster_from_files(
    files,
    classes: list = None,
    start: float = None,
    step: float = None,
    stop: float = None,
    extent_filter: BaseGeometry = None,
):
    """Create a new vectorized GeoData from a raster file."""
    with get_temporary_directory() as temp_dir:
        if len(files) != 1:
            raise UploadError("A raster data should use only one GeoTiff file")
        file_obj = files[0]
        file_name = secure_filename(file_obj.filename)
        file_path = os.path.join(temp_dir, file_name)
        file_obj.save(file_path)
        geo_data = load_from_geotiff(
            file_path,
            classes=classes,
            start=start,
            step=step,
            stop=stop,
            extent_filter=extent_filter,
        )
        return geo_data



"""Make MR-Sort explainability self-sufficient.

Revision ID: 5f84e56d0465f
Revises: a212e7950484
Create Date: 2025-04-01 14:45:00.000000

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "5f84e56d0465f"
down_revision = "a212e7950484"
branch_labels = None
depends_on = None


feature = sa.Table(
    "feature",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("feature", "geo_feature", name="featuretype"),
        nullable=True,
    ),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column(
        "explainability",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
)


def upgrade():
    conn = op.get_bind()
    # Get all geo features with non-polygon geometry
    nb = conn.execute(
        sa.text("SELECT count(id) FROM feature WHERE explainability != '[]'")
    ).fetchone()[0]
    res = conn.execute(
        sa.select(feature.c["id", "explainability"]).where(
            feature.c.explainability != []
        )
    )
    i = 0
    progress_bar(i, nb)
    for id, explain in res:
        if explain["model_type"] in ("mrsort", "geo_buffer"):
            conn.execute(
                feature.update()
                .values(explainability=[])
                .where(feature.c.id == id)
            )
        i += 1
        progress_bar(i, nb)


def downgrade():
    pass



from datetime import datetime, timedelta

import jwt

from .exceptions import InvalidAuthenticationHeader


class JWTHandler:
    secret_key = ""
    token_validity = 900
    signing_algorithm = "HS256"
    header_prefix = "Bearer"
    authentication_header = "Authorization"

    def init_app(self, app):
        self.app = app
        self.secret_key = app.config["API_TOKEN_SECRET_KEY"]
        self.token_validity = app.config.get(
            "API_TOKEN_VALIDITY", self.token_validity
        )
        self.refresh_validity = app.config.get("API_REFRESH_VALIDITY", None)
        self.signing_algorithm = app.config.get(
            "API_JWT_ALGORITHM", self.signing_algorithm
        )
        self.header_prefix = app.config.get(
            "API_JWT_HEADER_PREFIX", self.header_prefix
        )
        self.authentication_header = app.config.get(
            "API_JWT_AUTHENTICATION_HEADER", self.authentication_header
        )

    def create_base_payload(self):
        issue_time = datetime.utcnow()
        payload = {
            "iat": issue_time,
            "exp": issue_time + timedelta(seconds=self.token_validity),
        }
        return payload

    def create_token(self, payload):
        """Encode the payload in a JSON Web Token, and create
        the registered claims (eg. 'iat', 'exp')
        """
        base_payload = self.create_base_payload()
        payload.update(base_payload)  # overwrite the payload with base_payload
        return jwt.encode(
            payload, self.secret_key, algorithm=self.signing_algorithm
        )

    def create_user_token(self, user_id, payload=None):
        """Create a token for a user"""
        base_user_payload = {"sub": user_id}
        if payload is None:
            payload = base_user_payload
        else:
            payload.update(base_user_payload)
        return self.create_token(payload)

    def create_user_refresh_token(self, user_id):
        issue_time = datetime.utcnow()
        payload = {
            "sub": user_id,
            "iat": issue_time,
            "exp": issue_time + timedelta(seconds=self.refresh_validity),
        }
        return jwt.encode(
            payload, self.secret_key, algorithm=self.signing_algorithm
        )

    def get_token_exp(self, token):
        """Returns the expiration date of the token"""
        data = self.decode(token)
        return data.get("exp", None)

    def get_user_id(self, token):
        """Returns the user_id from a token"""
        data = self.decode(token)
        user_id = data.get("sub", None)
        return user_id

    def get_token_from_request(self, request):
        """Get the token from the flask request"""
        authorization = request.headers.get(self.authentication_header, None)
        if authorization is not None:
            try:
                auth_type, auth_cred = authorization.split(" ")
            except ValueError:
                raise InvalidAuthenticationHeader(
                    "Invalid JWT authentication header"
                )
            if auth_type == self.header_prefix:
                return auth_cred
            else:
                raise InvalidAuthenticationHeader(
                    "Incorrect authentication type for the JWT"
                )
        return None

    def encode(self, payload):
        """Encode the payload in a JSON Web Token"""
        return jwt.encode(
            payload, self.secret_key, algorithm=self.signing_algorithm
        )

    def decode(self, token):
        """Decode a JSON Web Token"""
        return jwt.decode(
            token, self.secret_key, algorithms=[self.signing_algorithm]
        )


jwt_handler = JWTHandler()
